"""Sphere

This module contains the Sphere class.
"""
# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #

from __future__ import division

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from microstructpy.geometry.n_sphere import NSphere

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Sphere Class                                                                #
#                                                                             #
# --------------------------------------------------------------------------- #
class Sphere(NSphere):
    """A 3D sphere.

    This class represents a three-dimensional circle. It is defined by
    a center point and size parameter, which can be either radius or diameter.

    Without input parameters, this defaults to a unit sphere centered at
    the origin.

    Args:
        r (float): *(optional)* The radius of the sphere.
            Defaults to 1.
        radius (float): *(optional)* Same as ``r``.
        d (float): *(optional)* Alias for ``2*r``.
        diameter (float): *(optional)* Alias for ``2*r``.
        size (float): *(optional)* Alias for ``2*r``.
        center (list, float, numpy.ndarray): *(optional)* The coordinates of
            the center. Defaults to ``[0, 0, 0]``.
        position (list, float, numpy.ndarray): *(optional)*
            Alias for ``center``.
    """
    # ----------------------------------------------------------------------- #
    # Constructor                                                             #
    # ----------------------------------------------------------------------- #
    def __init__(self, **kwargs):
        if 'volume' in kwargs:
            v = kwargs['volume']
            r = np.cbrt(3 * v / (4 * np.pi))
            kwargs['r'] = r

        NSphere.__init__(self, **kwargs)
        if len(self.center) == 0:
            self.center = tuple(self.n_dim * [0])

    # ----------------------------------------------------------------------- #
    # Representation Function                                                 #
    # ----------------------------------------------------------------------- #
    def __repr__(self):
        repr_str = 'Sphere('
        repr_str += 'center=' + repr(tuple(self.center))
        repr_str += ', radius=' + repr(self.r)
        repr_str += ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Number of Dimensions                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def n_dim(self):
        """int: number of dimensions, 3"""
        return 3

    # ----------------------------------------------------------------------- #
    # Volume                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def volume(self):
        """float: volume of sphere"""
        return 4 * np.pi * self.r * self.r * self.r / 3

    @classmethod
    def volume_expectation(cls, **kwargs):
        r"""Expected value of volume.

        This function computes the expected value for the volume of a sphere.
        The keyword arguments are identical to the :class:`.Sphere` function.
        The values for these keywords can be either constants or
        :mod:`scipy.stats` distributions.

        The expected value is computed by the following formula:

        .. math::

            \mathbb{E}[V] &= \mathbb{E}[\frac{4}{3}\pi R^3] \\
                            &= \frac{4}{3}\pi \mathbb{E}[R^3] \\
                            &= \frac{4}{3}\pi (\mu_R^3 + 3 \mu_R \sigma_R^2 + \gamma_{1, R} \sigma_R^3)

        Args:
            **kwargs: Keyword arguments, see :class:`microstructpy.geometry.Sphere`.

        Returns:
            float: Expected value of the volume of the sphere.

        """  # NOQA: E501
        # Check for radius distribution
        r_dist = None
        if 'radius' in kwargs:
            r_dist = kwargs['radius']
        elif 'r' in kwargs:
            r_dist = kwargs['r']

        if type(r_dist) in (float, int):
            return 4 * np.pi * r_dist * r_dist * r_dist / 3
        elif r_dist is not None:
            return 4 * np.pi * r_dist.moment(3) / 3

        # Check for diameter distribution
        d_dist = None
        for d_kw in ('d', 'diameter', 'size'):
            if d_kw in kwargs:
                d_dist = kwargs[d_kw]
                break

        if type(d_dist) in (float, int):
            return 0.5 * np.pi * d_dist * d_dist * d_dist / 3
        elif d_dist is not None:
            return 0.5 * np.pi * d_dist.moment(3) / 3

        if 'volume' in kwargs:
            v_dist = kwargs['volume']
            try:
                v_exp = v_dist.moment(1)
            except AttributeError:
                v_exp = v_dist

            return v_exp

        # Raise error
        e_str = 'Could not find one of the following keywords in the inputs: '
        e_str += 'r, radius, d, diameter, volume.'
        raise KeyError(e_str)

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the sphere.

        This function uses the :meth:`mpl_toolkits.mplot3d.Axes3D.plot_surface`
        method to add the sphere to the current axes. The keyword arguments
        are passed through to plot_surface.

        Args:
            **kwargs (dict): Keyword arguments for plot_surface.

        """  # NOQA: E501
        if len(plt.gcf().axes) == 0:
            ax = plt.axes(projection=Axes3D.name)
        else:
            ax = plt.gca()

        u = np.linspace(0, 2 * np.pi, 11)
        cv = np.linspace(-1, 1, 12)
        uu, cvv = np.meshgrid(u, cv)
        svv = np.sin(np.arccos(cvv))

        xc, yc, zc = self.center
        r = self.r

        xx = xc + r * np.cos(uu) * svv
        yy = yc + r * np.sin(uu) * svv
        zz = zc + r * cvv

        mod_kwargs = {}
        for key, val in kwargs.items():
            if key == 'facecolors' and type(val) != list:
                mod_kwargs['color'] = val
            else:
                mod_kwargs[key] = val
        ax.plot_surface(xx, yy, zz, **mod_kwargs)



import os

import numpy as np
import scipy.stats
from matplotlib import pyplot as plt

import microstructpy as msp


def main():
    # Create Directory
    dirname = os.path.join(os.path.dirname(__file__), 'foam')
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # Define Domain
    domain = msp.geometry.Square(side_length=8)

    # Create Void Tessellation
    void_mat = {'material_type': 'void',
                'shape': 'circle',
                'size': scipy.stats.lognorm(scale=1, s=0.2)
                }

    void_a = 0.7 * domain.area
    void_seeds = msp.seeding.SeedList.from_info(void_mat, void_a)
    void_seeds.position(domain, rtol=0.03, verbose=True)
    void_tess = msp.meshing.PolyMesh.from_seeds(void_seeds, domain)

    # Add Foam
    foam_mat = {'material_type': 'amorphous',
                'shape': 'circle',
                'size': scipy.stats.lognorm(scale=0.15, s=0.1)
                }

    foam_a = 0.15 * domain.area
    foam_seeds = msp.seeding.SeedList.from_info(foam_mat, foam_a)
    inds = np.flip(np.argsort([s.volume for s in foam_seeds]))
    foam_seeds = foam_seeds[inds]

    bkdwns = np.array([s.breakdown[0] for s in foam_seeds])
    np.random.seed(0)
    for i, seed in enumerate(foam_seeds):
        if i == 0:
            trial_pt = trial_position(void_tess)
        else:
            r = seed.geometry.r
            check_bkdwns = bkdwns[:i]
            good_pt = False
            while not good_pt:
                trial_pt = trial_position(void_tess)
                good_pt = check_pt(trial_pt, r, check_bkdwns)

        seed.position = trial_pt
        bkdwns[i] = seed.breakdown
        seed.phase = 1

    # Combine Results
    materials = [void_mat, foam_mat]
    seeds = void_seeds + foam_seeds
    pmesh = msp.meshing.PolyMesh.from_seeds(seeds, domain)

    # Triangular Mesh
    tmesh = msp.meshing.TriMesh.from_polymesh(pmesh,
                                              materials,
                                              min_angle=20,
                                              max_edge_length=0.1)

    # Plot
    tmesh.plot(facecolor='aquamarine',
               edgecolor='k',
               linewidth=0.2)

    plt.gca().set_position([0, 0, 1, 1])
    plt.axis('image')
    plt.gca().set_axis_off()
    plt.gca().get_xaxis().set_visible(False)
    plt.gca().get_yaxis().set_visible(False)

    xlim, ylim = domain.limits
    plt.axis([xlim[0], xlim[1], ylim[0], ylim[1]])

    for ext in ['png', 'pdf']:
        fname = os.path.join(dirname, 'trimesh.' + ext)
        plt.savefig(fname, bbox_inches='tight', pad_inches=0)


def pick_edge(void_tess):
    f_neighs = void_tess.facet_neighbors
    i = -1
    neighs = [-1, -1]
    while any([n < 0 for n in neighs]):
        i = np.random.randint(len(f_neighs))
        neighs = f_neighs[i]
    facet = void_tess.facets[i]
    j = np.random.randint(len(facet))
    kp1 = facet[j]
    kp2 = facet[j - 1]
    return kp1, kp2


def trial_position(void_tess):
    kp1, kp2 = pick_edge(void_tess)
    pt1 = void_tess.points[kp1]
    pt2 = void_tess.points[kp2]

    f = np.random.rand()
    return [f * x1 + (1 - f) * x2 for x1, x2 in zip(pt1, pt2)]


def check_pt(point, r, breakdowns):
    pts = breakdowns[:, :-1]
    rads = breakdowns[:, -1]

    rel_pos = pts - point
    dist = np.sqrt(np.sum(rel_pos * rel_pos, axis=1))
    min_dist = rads + r - 0.3 * np.minimum(rads, r)
    return np.all(dist > min_dist)


if __name__ == '__main__':
    main()



# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
from __future__ import division

import numpy as np
from matplotlib import patches
from matplotlib import pyplot as plt

from microstructpy.geometry.n_box import NBox

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Rectangle Class                                                             #
#                                                                             #
# --------------------------------------------------------------------------- #
class Rectangle(NBox):
    """Rectangle

    This class contains a generic, 2D rectangle. The position and dimensions
    of the box can be specified using any of the parameters below.

    Without parameters, this returns a unit square centered on the origin.

    Args:
        length (float): *(optional)* Length of the rectangle.
        width (float): *(optional)* Width of the rectangle. *(optional)*
        side_lengths (list): *(optional)* Side lengths. Defaults to (1, 1).
        center (list): *(optional)* Center of rectangle. Defaults to (0, 0).
        corner (list): *(optional)* Bottom-left corner.
        bounds (list): *(optional)* Bounds of rectangle. Expected to be in the
            format [(xmin, xmax), (ymin, ymax)].
        limits : Alias for *bounds*.
        angle (float): *(optional)* The rotation angle, in degrees.
        angle_deg (float): *(optional)* The rotation angle, in degrees.
        angle_rad (float): *(optional)* The rotation angle, in radians.
        matrix (numpy.ndarray): *(optional)* The 2x2 rotation matrix.
    """

    def __init__(self, **kwargs):
        if 'length' in kwargs and 'width' in kwargs:
            kwargs['side_lengths'] = [kwargs['length'], kwargs['width']]

        if 'angle' in kwargs:
            cp = np.cos(np.radians(kwargs['angle']))
            sp = np.sin(np.radians(kwargs['angle']))
            kwargs['matrix'] = np.array([[cp, -sp], [sp, cp]])
        elif 'angle_deg' in kwargs:
            cp = np.cos(np.radians(kwargs['angle_deg']))
            sp = np.sin(np.radians(kwargs['angle_deg']))
            kwargs['matrix'] = np.array([[cp, -sp], [sp, cp]])
        elif 'angle_rad' in kwargs:
            cp = np.cos(kwargs['angle_rad'])
            sp = np.sin(kwargs['angle_rad'])
            kwargs['matrix'] = np.array([[cp, -sp], [sp, cp]])

        NBox.__init__(self, **kwargs)
        try:
            self.center
        except AttributeError:
            self.center = [0, 0]

        try:
            self.side_lengths
        except AttributeError:
            self.side_lengths = [1, 1]

    # ----------------------------------------------------------------------- #
    # Best Fit Function                                                       #
    # ----------------------------------------------------------------------- #
    def best_fit(self, points):
        """Find rectangle of best fit for points

        Args:
            points (list): List of points to fit.

        Returns:
            Rectangle: an instance of the class that best fits the points.
        """
        # Unpack the input points
        pts = np.array(points, dtype='float')
        x, y = pts.T

        # Find the most likely orientation for the rectangle
        A = np.vstack([x, np.ones(len(x))]).T
        m, _ = np.linalg.lstsq(A, y, rcond=None)[0]
        theta = np.arctan(m)

        # Rotate the points to an axis-aligned frame
        s = np.sin(theta)
        c = np.cos(theta)
        rot = np.array([[c, -s], [s, c]])
        pts_prin = pts.dot(rot)

        # Translate points to center of bounding box
        mins = pts_prin.min(axis=0)
        maxs = pts_prin.max(axis=0)
        bb_cen = 0.5 * (mins + maxs)
        pts_bb = pts_prin - bb_cen

        x_min, y_min = pts_bb.min(axis=0)
        x_max, y_max = pts_bb.max(axis=0)

        # Find nearest edge
        x1_dist = np.abs(pts_bb[:, 0] - x_min)
        x2_dist = np.abs(pts_bb[:, 0] - x_max)
        y1_dist = np.abs(pts_bb[:, 1] - y_min)
        y2_dist = np.abs(pts_bb[:, 1] - y_max)

        dists = np.array([x1_dist, x2_dist, y1_dist, y2_dist]).T
        min_dist = dists.min(axis=1)

        mask_x1 = np.isclose(x1_dist, min_dist)
        mask_x2 = np.isclose(x2_dist, min_dist)
        mask_y1 = np.isclose(y1_dist, min_dist)
        mask_y2 = np.isclose(y2_dist, min_dist)

        # Group points by edge
        x_x1 = pts_bb[mask_x1, 0].T
        y_y1 = pts_bb[mask_y1, 1].T
        x_x2 = pts_bb[mask_x2, 0].T
        y_y2 = pts_bb[mask_y2, 1].T

        # Get lines of best fit for each edge
        x1_fit = np.mean(x_x1)
        y1_fit = np.mean(y_y1)
        x2_fit = np.mean(x_x2)
        y2_fit = np.mean(y_y2)

        x_cen = 0.5 * (x1_fit + x2_fit)
        y_cen = 0.5 * (y1_fit + y2_fit)
        fit_cen_bb = np.array([x_cen, y_cen])
        x_len = x2_fit - x1_fit
        y_len = y2_fit - y1_fit

        # Translate center to rotated frame
        fit_cen_prin = fit_cen_bb + bb_cen
        fit_cen = rot.dot(fit_cen_prin.reshape(-1, 1)).flatten()

        # Disambiguate the orientation and axes
        x_ax_seed = np.array(self.matrix)[:, 0]
        x_dot, y_dot = rot.T.dot(x_ax_seed)

        if np.abs(x_dot) > np.abs(y_dot):
            if x_dot > 0:
                x_ax_fit = rot[:, 0]
            else:
                x_ax_fit = - rot[:, 0]

            length = x_len
            width = y_len
        else:
            if y_dot > 0:
                x_ax_fit = rot[:, 1]
            else:
                x_ax_fit = - rot[:, 1]

            length = y_len
            width = x_len

        ang_diff = np.arcsin(np.cross(x_ax_seed, x_ax_fit))
        ang_rad = self.angle_rad + ang_diff

        rect_fit = Rectangle(center=fit_cen, length=length, width=width,
                             angle_rad=ang_rad)

        return rect_fit

    # ----------------------------------------------------------------------- #
    # Representation Function                                                 #
    # ----------------------------------------------------------------------- #
    def __repr__(self):
        repr_str = 'Rectangle('
        repr_str += 'center=' + repr(tuple(self.center)) + ', '
        repr_str += 'side_lengths=' + repr(tuple(self.side_lengths)) + ', '
        repr_str += 'angle=' + repr(self.angle) + ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Number of Dimensions                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def n_dim(self):
        """int: Number of dimensions, 2"""
        return 2

    # ----------------------------------------------------------------------- #
    # Area                                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def area(self):
        """float: Area of rectangle"""
        return self.n_vol

    @classmethod
    def area_expectation(cls, **kwargs):
        r"""Expected area of rectangle

        This method computes the expected area of a rectangle. There are two
        main ways to define the size of a rectangle: by the length and width
        and by the bounds. If the input rectangle is defined by length and
        width, the expected area is:

        .. math::

            \mathbb{E}[A] = \mathbb{E}[L W] = \mu_L \mu_W

        For the case where it is defined by upper and lower bounds:

        .. math::

            \mathbb{E}[A] = \mathbb{E}[(X_U - X_L) (Y_U - Y_L)]

        .. math::
            \mathbb{E}[A] =
            \mu_{X_U}\mu_{Y_U} - \mu_{X_U} \mu_{Y_L} -
            \mu_{X_L}\mu_{Y_U} + \mu_{X_L}\mu_{Y_L}

        Example:
            >>> import scipy.stats
            >>> import microstructpy as msp
            >>> L = scipy.stats.uniform(loc=1, scale=2)
            >>> W = scipy.stats.norm(loc=3.2, scale=5.1)
            >>> L.mean() * W.mean()
            6.4
            >>> msp.geometry.Rectangle.area_expectation(length=L, width=W)
            6.4

        Args:
            **kwargs: Keyword arguments, same as :class:`.Rectangle` but the
                inputs can be from the :mod:`scipy.stats` module.

        Returns:
            float: Expected/average area of rectangle.

        """
        if 'length' in kwargs or 'width' in kwargs:
            len_dist = kwargs.get('length', 1)
            width_dist = kwargs.get('width', 1)
            return _prod_exp(*[len_dist, width_dist])

        if 'side_lengths' in kwargs:
            return _prod_exp(*kwargs['side_lengths'])

        if 'bounds' in kwargs:
            x_bnds, y_bnds = kwargs['bounds']
            x_lb, x_ub = x_bnds
            y_lb, y_ub = y_bnds

            p1 = _prod_exp(*[x_ub, y_ub])
            p2 = _prod_exp(*[x_ub, y_lb])
            p3 = _prod_exp(*[x_lb, y_ub])
            p4 = _prod_exp(*[x_lb, y_lb])

            return p1 - p2 - p3 + p4

        raise ValueError('Cannot find the expected area of rectangle')

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #
    @property
    def length(self):
        """float: Length of rectangle, side length along 1st axis"""
        return self.side_lengths[0]

    @property
    def width(self):
        """float: Width of rectangle, side length along 2nd axis"""
        return self.side_lengths[1]

    @property
    def angle(self):
        """float: Rotation angle of rectangle - degrees"""
        return self.angle_deg

    @property
    def angle_deg(self):
        """float: Rotation angle of rectangle - degrees"""
        return 180 * self.angle_rad / np.pi

    @property
    def angle_rad(self):
        """float: Rotation angle of rectangle - radians"""
        return np.arctan2(self.matrix[1][0], self.matrix[0][0])

    # ----------------------------------------------------------------------- #
    # Circle Approximation                                                    #
    # ----------------------------------------------------------------------- #
    def approximate(self, x1=None):
        """Approximate rectangle with a set of circles.

        This method approximates a rectangle with a set of circles.
        These circles are spaced uniformly along the long axis of the
        rectangle with distance ``x1`` between them.

        Example
        -------

        For a rectangle with length=2.5, width=1, and x1=0.3,
        the approximation would look like :numref:`f_api_rectangle_approx`.

        .. _f_api_rectangle_approx:
        .. figure:: ../../auto_examples/geometry/images/sphx_glr_plot_rectangle_001.png

            Circular approximation of rectangle.

        Args:
            x1 (float or None): *(optional)* Spacing between the circles.
                If not specified, the spacing is 0.25x the length of the
                shortest side.

        Returns:
            numpy.ndarray: An Nx3 array, where each row is a circle and the
            columns are x, y, and r.

        """  # NOQA: E501
        if x1 is None:
            x1 = 0.25 * min(self.side_lengths)

        if self.side_lengths[0] >= self.side_lengths[1]:
            length = self.side_lengths[0]
            width = self.side_lengths[1]
            inds = [0, 1]
        else:
            length = self.side_lengths[1]
            width = self.side_lengths[0]
            inds = [1, 0]

        # Centerline circles
        xc = 0
        half_len = 0.5 * length
        r = 0.5 * width
        circs = []
        while xc < half_len:
            circ = [xc, 0, r]
            circs.append(circ)

            if np.isclose(xc + r, half_len):
                xc = half_len
            elif xc + x1 > half_len - r:
                xc = half_len - r
            else:
                xc = xc + x1

        # Corner circle
        while r > 0:
            x_init = circs[-1][0]
            y_init = circs[-1][1]
            x = x_init + x1
            y = y_init + x1

            dx = half_len - x
            dy = 0.5 * width - y
            r = min(dx, dy)
            if r <= 0:
                break
            else:
                circs.append([x, y, r])

        # Reflect circles
        circs = np.array(circs)
        for dim in range(self.n_dim):
            mask = circs[:, dim] > 0
            new_circs = np.copy(circs[mask])
            new_circs[:, dim] *= -1
            circs = np.concatenate((circs, new_circs))

        circs[:, inds] = circs[:, [0, 1]]

        # Rotate and translate circles
        pts = circs[:, :-1]
        circs[:, :-1] = pts.dot(np.array(self.matrix).T)
        circs[:, :-1] += self.center
        return circs

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the rectangle.

        This function adds a :class:`matplotlib.patches.Rectangle` patch to the
        current axes. The keyword arguments are passed through to the patch.

        Args:
            **kwargs (dict): Keyword arguments for the patch.

        """  # NOQA: E501
        pt = self.corner
        w = self.length
        h = self.width
        ang = self.angle
        c = patches.Rectangle(pt, w, h, ang, **kwargs)
        plt.gca().add_patch(c)


def _prod_exp(*args):
    prod = 1
    for arg in args:
        try:
            arg_mu = arg.moment(1)
        except AttributeError:
            arg_mu = arg
        prod *= arg_mu
    return prod


# --------------------------------------------------------------------------- #
#                                                                             #
# Square Class                                                                #
#                                                                             #
# --------------------------------------------------------------------------- #
class Square(Rectangle):
    """A square.

    This class contains a generic, 2D square. It is derived from the
    :class:`microstructpy.geometry.Rectangle` class and contains the
    ``side_length`` property, rather than multiple side lengths.

    Args:
        side_length (float): *(optional)* Side length. Defaults to 1.
        center (list): *(optional)* Center of rectangle. Defaults to (0, 0).
        corner (list): *(optional)* Bottom-left corner.
    """

    def __init__(self, **kwargs):
        if 'side_length' in kwargs:
            kwargs['side_lengths'] = 2 * [kwargs['side_length']]

        Rectangle.__init__(self, **kwargs)

    # ----------------------------------------------------------------------- #
    # Side Length Property                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def side_length(self):
        """float: length of the side of the square."""
        return self.side_lengths[0]

    # ----------------------------------------------------------------------- #
    # Area                                                                    #
    # ----------------------------------------------------------------------- #
    @classmethod
    def area_expectation(cls, **kwargs):
        r"""Expected area of square

        This method computes the expected area of a square with distributed
        side length.
        The expectation is:

        .. math::

            \mathbb{E}[A] = \mathbb{E}[S^2] = \mu_S^2 + \sigma_S^2

        Example:
            >>> import scipy.stats
            >>> import microstructpy as msp
            >>> S = scipy.stats.expon(scale=2)
            >>> S.mean()^2 + S.var()
            8.0
            >>> msp.geometry.Square.area_expectation(side_length=S)
            8.0

        Args:
            **kwargs: Keyword arguments, same as :class:`.Square` but the
                inputs can be from the :mod:`scipy.stats` module.

        Returns:
            float: Expected/average area of the square.

        """
        if 'side_length' in kwargs:
            len_dist = kwargs['side_length']

            try:
                area_exp = len_dist.moment(2)
            except AttributeError:
                area_exp = len_dist * len_dist
            return area_exp

        Rectangle.area_expectation(**kwargs)

    # ----------------------------------------------------------------------- #
    # Circle Approximation                                                    #
    # ----------------------------------------------------------------------- #
    def approximate(self, x1=None):
        """Approximate square with a set of circles

        This method approximates a square with a set of circles.
        These circles are spaced uniformly along the edges of the square
        with distance ``x1`` between them.

        Example
        -------

        For a square with side_length=1, and x1=0.2,
        the approximation would look like :numref:`f_api_square_approx`.

        .. _f_api_square_approx:
        .. figure:: ../../auto_examples/geometry/images/sphx_glr_plot_rectangle_002.png

            Circular approximation of square.

        Args:
            x1 (float or None): *(optional)* Spacing between the circles.
                If not specified, the spacing is 0.25x the side length.

        Returns:
            numpy.ndarray: An Nx3 array, where each row is a circle and the
            columns are x, y, and r.

        """  # NOQA: E501
        return Rectangle.approximate(self, x1)



"""Miscellaneous functions

This private module contains miscellaneous functions.
"""
import numpy as np

__author__ = 'Kenneth (Kip) Hart'

kw_solid = {'crystalline', 'granular', 'solid'}
kw_amorph = {'amorphous', 'glass', 'matrix'}
kw_void = {'void', 'crack', 'hole'}

ori_kws = {'orientation', 'matrix', 'angle', 'angle_deg', 'angle_rad',
           'rot_seq', 'rot_seq_rad', 'rot_seq_deg'}
gen_kws = {'material_type', 'fraction', 'shape', 'name', 'color', 'position'}

demo_needs = {'basalt_circle.xml': ['aphanitic_cdf.csv', 'olivine_cdf.csv'],
              'from_image.py': ['aluminum_micro.png']}

mpl_plural_kwargs = {'edgecolors', 'facecolors', 'linewidths', 'antialiaseds',
                     'offsets'}
plt_3d_adj = {
    'left': 0.4,
    'right': 1,
    'bottom': 0,
    'top': 0.8,
}


# --------------------------------------------------------------------------- #
#                                                                             #
# Convert String to Value (Infer Type)                                        #
#                                                                             #
# --------------------------------------------------------------------------- #
def from_str(string):
    """ Convert string to number

    This function takes a string and converts it into a number or a list.

    Args:
        string (str): The string.

    Returns:
        The value in the string.

    """
    beg_delims = ('(', '[', '{', '<')
    end_delims = (')', ']', '}', '>')

    string = string.strip()
    if any([c in string for c in beg_delims + end_delims]) or ',' in string:
        if string[0] in beg_delims:
            string = string[1:]
        if string[-1] in end_delims:
            string = string[:-1]
        val = []
        n_beg = 0
        n_end = 0
        elem_str = ''
        for char in string:
            if char in beg_delims:
                n_beg += 1
            elif char in end_delims:
                n_end += 1

            if (char == ',') and n_beg == n_end:
                val.append(from_str(elem_str.strip()))
                elem_str = ''
            else:
                elem_str += char
        val.append(from_str(elem_str.strip()))
        return val
    else:
        try:
            val = int(string)
        except ValueError:
            try:
                val = float(string)
            except ValueError:
                if string.lower() in ('true', 'yes'):
                    val = True
                elif string.lower() in ('false', 'no'):
                    val = False
                else:
                    val = str(string)
        return val


# --------------------------------------------------------------------------- #
#                                                                             #
# Tangent Spheres                                                             #
#                                                                             #
# --------------------------------------------------------------------------- #
def tangent_sphere(points, radii=None, simplices=None):
    """Calculate center and radius of tangent sphere(s)

    This function computes the center and radius of an n-dimensional sphere
    that is tangent to (n+1) spheres. For example, in 2D this function
    computes the center and radius of a circle tangent to three other circles.

    The operation of this function can be vectorized using the ``simplices``
    input. The simplices should be an Mx(n+1) list of indices of the points.
    The result is an Mx(n+1) numpy array, where the first n columns are the
    coordinates of the sphere center. The final column is the radius of the
    sphere.

    If no radii are specified, the results are circumspheres of the simplices
    (circumcircles in 2D).

    Radii at each point can be speficied. If no radii are given, then the
    results are circumspheres of the simplices (circumcircles in 2D).

    Args:
        points (list, tuple, numpy.ndarray): List of points.
        radii (list, tuple, numpy.ndarray): List of radii. *(optional)*
        simplices (list, tuple, numpy.ndarray): List of simplices. *(optional)*

    Returns:
        numpy.ndarray: The centers and radii of tangent spheres.

    """
    # set radii
    if radii is None:
        radii = np.full(len(points), 0)

    # extract points
    if simplices is None:
        simplices = np.arange(len(points)).reshape(1, -1)

    pts = np.array(points)[simplices]
    rs = np.array(radii)[simplices]

    # define circle distances
    cs = np.sum(pts * pts, axis=-1) - rs * rs

    # matrix and vector quantities
    pos1 = pts[:, 0]
    r1 = rs[:, 0]
    A = pts[:, 1:] - pos1[:, np.newaxis, :]
    b = -1 * (rs[:, 1:] - r1[:, np.newaxis])
    c = 0.5 * (cs[:, 1:] - cs[:, 0, np.newaxis])

    # linear system coefficients
    alpha = np.linalg.solve(A, b)
    beta = np.linalg.solve(A, c)

    # quadratic equation in rc
    r_beta = beta - pos1
    C1 = np.sum(alpha * alpha, axis=-1) - 1
    C2 = np.sum(r_beta * alpha, axis=-1) - r1
    C3 = np.sum(r_beta * r_beta, axis=-1) - r1 * r1

    # solve for rc
    discr = C2 * C2 - C1 * C3
    rt_discr = np.sqrt(discr)
    rt_discr[discr < 0] = 0

    rc1 = (-C2 + rt_discr) / C1
    rc2 = (-C2 - rt_discr) / C1

    mask = np.abs(rc1) < np.abs(rc2)
    rc = rc2
    rc[mask] = rc1[mask]
    rc[discr < 0] = 0

    # solve for center position
    posc = alpha * rc[:, np.newaxis] + beta

    # return results
    spheres = np.hstack((posc, rc.reshape(-1, 1)))
    return np.squeeze(spheres)


def axisEqual3D(ax):
    '''From stackoverflow: https://stackoverflow.com/a/19248731'''
    extents = np.array([getattr(ax, 'get_{}lim'.format(d))() for d in 'xyz'])
    sz = extents[:, 1] - extents[:, 0]
    centers = np.mean(extents, axis=1)
    maxsize = max(abs(sz))
    r = maxsize/2
    for ctr, dim in zip(centers, 'xyz'):
        getattr(ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)


def ax_objects(ax):
    n = 0
    for att in ['collections', 'images', 'lines', 'patches', 'texts']:
        n += len(getattr(ax, att))
    return n



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('../../src/'))
sys.path.insert(0, os.path.abspath('../../src/MicroStructPy/'))

import sphinx_gallery

import microstructpy


# -- Project information -----------------------------------------------------

project = 'MicroStructPy'
copyright = '2020, Georgia Tech Research Corporation'
author = 'Kenneth Hart'

# The short X.Y version
version = ''.join(microstructpy.__version__.split('.')[:2])
# The full version, including alpha/beta/rc tags
release = microstructpy.__version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx_gallery.gen_gallery',
]

intersphinx_mapping = {
    'matplotlib': ('https://matplotlib.org/', None),
    'meshpy': ('https://documen.tician.de/meshpy/', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'python': ('https://docs.python.org/3', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None)
}
# Suppress "WARNING: unknown mimetype for ..."
suppress_warnings = ['epub.unknown_project_files']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ['auto_examples/*', 'sphinx_gallery/*']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# Mock imports
autodoc_mock_imports = [
    'aabbtree',
    'matplotlib',
    'meshpy',
    'mpl_toolkits',
    'numpy',
    'pyquaternion',
    'pyvoro',
    'pyvtk',
    'scipy',
    'sphinx-gallery',
    'xmltodict']
autodoc_member_order = 'groupwise'

sphinx_gallery_conf = {
    'examples_dirs': [
        'sphinx_gallery',
        'sphinx_gallery/geometry',
        ],
    'gallery_dirs': [
        'auto_examples/demos',
        'auto_examples/geometry',
        ],
}

# Numbering figures, tables, listings, etc
numfig = True
numfig_secnum_depth = 1



# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'logo': 'logo/pad_logo.svg',
    'logo_name': True,
    'sidebar_width': '230px',
    'description': 'Microstructure modeling, mesh generation, analysis, and visualization.',
    'analytics_id': 'UA-147258715-1',
    'code_font_size': '12.5px',
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'MicroStructPy'

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = 'MSP'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = '_static/logo.svg'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = False

# If false, no index is generated.
html_use_index = False

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'MicroStructPydoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    'preamble': '\n'.join(['\\usepackage{subfig}',
                           '\\numberwithin{figure}{chapter}',
                           '\\numberwithin{table}{chapter}',
                           '\\numberwithin{equation}{chapter}'
                           ]),

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

latex_logo = '_static/logo/pad_logo.png'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('contents', 'MicroStructPy.tex', 'MicroStructPy Documentation',
     'Kenneth Hart', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'microstructpy', 'MicroStructPy Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'MicroStructPy', 'MicroStructPy Documentation',
     author, 'MicroStructPy',
     'Microstructure modeling, mesh generation, analysis, and visualization.',
     'Miscellaneous'),
]



"""Triangle/Tetrahedron Meshing

This module contains the class definition for the TriMesh class.

"""
# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #


from __future__ import division
from __future__ import print_function

import meshpy.tet
import meshpy.triangle
import numpy as np
from matplotlib import collections
from matplotlib import patches
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from microstructpy import _misc

__all__ = ['TriMesh']
__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# TriMesh Class                                                               #
#                                                                             #
# --------------------------------------------------------------------------- #
class TriMesh(object):
    """Triangle/Tetrahedron mesh.

    The TriMesh class contains the points, facets, and elements in a triangle/
    tetrahedron mesh, also called an unstructured grid.

    The points attribute is an Nx2 or Nx3 list of points in the mesh.
    The elements attribute contains the Nx3 or Nx4 list of the points at
    the corners of each triangle/tetrahedron. A list of facets can also be
    included, though it is optional and does not need to include every facet
    in the mesh. Attributes can also be assigned to the elements and facets,
    though they are also optional.

    Args:
        points (list, numpy.ndarray): List of coordinates in the mesh.
        elements (list, numpy.ndarray): List of indices of the points at
            the corners of each element. The shape should be Nx3 in 2D or
            Nx4 in 3D.
        element_attributes (list, numpy.ndarray): *(optional)* A number
            associated with each element.
            Defaults to None.
        facets (list, numpy.ndarray): *(optional)* A list of facets in the
            mesh. The shape should be Nx2 in 2D or Nx3 in 3D.
            Defaults to None.
        facet_attributes (list, numpy.ndarray): *(optional)* A number
            associated with each facet.
            Defaults to None.

    """
    # ----------------------------------------------------------------------- #
    # Constructors                                                            #
    # ----------------------------------------------------------------------- #
    def __init__(self, points, elements, element_attributes=None, facets=None,
                 facet_attributes=None):
        self.points = points
        self.elements = elements
        self.element_attributes = element_attributes
        self.facets = facets
        self.facet_attributes = facet_attributes

    @classmethod
    def from_file(cls, filename):
        """Read TriMesh from file.

        This function reads in a triangular mesh from a file and creates an
        instance from that file. Currently the only supported file type
        is the output from :meth:`.write` with the ``format='str'`` option.

        Args:
            filename (str): Name of file to read from.

        Returns:
            TriMesh: An instance of the class.

        """
        with open(filename, 'r') as file:
            stage = 0
            pts = []
            elems = []
            elem_atts = []
            facets = []
            facet_atts = []

            n_eas = 0
            n_facets = 0
            n_fas = 0
            for line in file.readlines():
                if 'Mesh Points'.lower() in line.lower():
                    n_pts = int(line.split(':')[1])
                    stage = 'points'
                elif 'Mesh Elements'.lower() in line.lower():
                    n_elems = int(line.split(':')[1])
                    stage = 'elements'
                elif 'Element Attributes'.lower() in line.lower():
                    n_eas = int(line.split(':')[1])
                    stage = 'element attributes'
                elif 'Facets'.lower() in line.lower():
                    n_facets = int(line.split(':')[1])
                    stage = 'facets'
                elif 'Facet Attributes'.lower() in line.lower():
                    n_fas = int(line.split(':')[1])
                    stage = 'facet attributes'
                else:
                    if stage == 'points':
                        pts.append([float(x) for x in line.split(',')])
                    elif stage == 'elements':
                        elems.append([int(kp) for kp in line.split(',')])
                    elif stage == 'element attributes':
                        elem_atts.append(_misc.from_str(line))
                    elif stage == 'facets':
                        facets.append([int(kp) for kp in line.split(',')])
                    elif stage == 'facet attributes':
                        facet_atts.append(_misc.from_str(line))
                    else:
                        pass

        # check the inputs
        assert len(pts) == n_pts
        assert len(elems) == n_elems
        assert len(elem_atts) == n_eas
        assert len(facets) == n_facets
        assert len(facet_atts) == n_fas

        return cls(pts, elems, elem_atts, facets, facet_atts)

    @classmethod
    def from_polymesh(cls, polymesh, phases=None, min_angle=0,
                      max_volume=float('inf'), max_edge_length=float('inf')):
        """Create TriMesh from PolyMesh.

        This constuctor creates a triangle/tetrahedron mesh from a polygon
        mesh (:class:`.PolyMesh`). Polygons of the same seed number are
        merged and the element attribute is set to the seed number it is
        within. The facets between seeds are saved to the mesh and the index
        of the facet is stored in the facet attributes.

        Since the PolyMesh can include phase numbers for each region,
        additional information about the phases can be included as an input.
        The "phases" input should be a list of material phase dictionaries,
        formatted according to the :ref:`phase_dict_guide` guide.

        The minimum angle, maximum volume, and maximum edge length options
        provide quality controls for the mesh. The phase type option can take
        one of several values, described below.

        * **crystalline**: granular, solid
        * **amorphous**: glass, matrix
        * **void**: crack, hole

        The **crystalline** option creates a mesh where cells of the same seed
        number are merged, but cells are not merged across seeds. _This is
        the default material type._

        The **amorphous** option creates a mesh where cells of the same
        phase number are merged to create an amorphous region in the mesh.

        Finally, the **void** option will merge neighboring void cells and
        treat them as holes in the mesh.

        Args:
            polymesh (PolyMesh): A polygon/polyhedron mesh.
            phases (list): *(optional)* A list of dictionaries containing
                options for each phase.
                Default is
                ``{'material_type': 'solid', 'max_volume': float('inf')}``.
            min_angle (float): The minimum interior angle of an element.
            max_volume (float): The default maximum cell volume, used if one
                is not set for each phase.
            max_edge_length (float): The maximum edge length of elements
                along grain boundaries. Currently only supported in 2D.

        """
        # condition the phases input
        if phases is None:
            default_dict = {'material_type': 'solid',
                            'max_volume': float('inf')}
            n_phases = int(np.max(polymesh.phase_numbers)) + 1
            phases = [default_dict for _ in range(n_phases)]

        # create point and facet lists
        kps = {}
        pts = []
        facets = []
        facet_neighs = []
        facet_nums = []
        for i in range(len(polymesh.facets)):
            facet = polymesh.facets[i]
            neighs = polymesh.facet_neighbors[i]
            if facet_check(neighs, polymesh, phases):
                new_facet = []
                for kp_old in facet:
                    if kp_old not in kps:
                        kp_new = len(pts)
                        pts.append(polymesh.points[kp_old])
                        kps[kp_old] = kp_new
                    else:
                        kp_new = kps[kp_old]
                    new_facet.append(kp_new)
                facets.append(new_facet)
                facet_neighs.append(neighs)
                facet_nums.append(i + 1)

        # Subdivide facets
        n_dim = len(pts[0])
        if n_dim == 2:
            n_subs = np.ones(len(facets), dtype='int')
            for i, facet in enumerate(facets):
                pt1 = np.array(pts[facet[0]])
                pt2 = np.array(pts[facet[1]])
                rel_pos = pt2 - pt1
                n_float = np.linalg.norm(rel_pos) / max_edge_length
                n_int = max(1, np.ceil(n_float))
                n_subs[i] = n_int
            sub_out = meshpy.triangle.subdivide_facets(n_subs, pts, facets,
                                                       facet_nums)
            pts, facets, facet_nums = sub_out

        # create groups/regions
        pts_arr = np.array(polymesh.points)
        regions = []
        holes = []

        ungrouped = np.full(len(polymesh.regions), True, dtype='?')
        while np.any(ungrouped):
            cell_ind = np.argmax(ungrouped)

            # compute cell center
            facet_list = polymesh.regions[cell_ind]
            cell_kps = set()
            [cell_kps.update(polymesh.facets[n]) for n in facet_list]
            cell_cen = pts_arr[list(cell_kps)].mean(axis=0)

            # seed number and phase type
            seed_num = int(polymesh.seed_numbers[cell_ind])
            phase_num = polymesh.phase_numbers[cell_ind]
            phase = phases[phase_num]
            phase_type = phase.get('material_type', 'crystalline')
            phase_vol = phase.get('max_volume', max_volume)

            # get all cell numbers in group
            cell_nums = set([cell_ind])
            old_len = len(cell_nums)
            searching_front = True
            while searching_front:
                front = set()
                for n in cell_nums:
                    neighs = set()
                    for facet_num in polymesh.regions[n]:
                        f_neighs = polymesh.facet_neighbors[facet_num]
                        neigh_ind = [i for i in f_neighs if i != n][0]
                        if neigh_ind < 0:
                            continue
                        if not facet_check(f_neighs, polymesh, phases):
                            neighs.add(neigh_ind)
                    assert ungrouped[list(neighs)].all()
                    front.update(neighs)
                cell_nums |= front
                new_len = len(cell_nums)
                searching_front = new_len != old_len
                old_len = new_len

            ungrouped[list(cell_nums)] = False

            # update appropriate list
            if phase_type in _misc.kw_void:
                holes.append(cell_cen)
            else:
                regions.append(cell_cen.tolist() + [seed_num, phase_vol])

        # build inputs
        if n_dim == 2:
            info = meshpy.triangle.MeshInfo()
        else:
            info = meshpy.tet.MeshInfo()

        info.set_points(pts)
        info.set_facets(facets, facet_nums)
        info.set_holes(holes)

        info.regions.resize(len(regions))
        for i, r in enumerate(regions):
            info.regions[i] = tuple(r)

        # run MeshPy
        if n_dim == 2:
            tri_mesh = meshpy.triangle.build(info,
                                             attributes=True,
                                             volume_constraints=True,
                                             max_volume=max_volume,
                                             min_angle=min_angle,
                                             generate_faces=True)
        else:
            opts = meshpy.tet.Options('pq')
            opts.mindihedral = min_angle
            opts.maxvolume = max_volume
            opts.fixedvolume = 1
            opts.regionattrib = 1
            opts.facesout = 1
            tri_mesh = meshpy.tet.build(info, options=opts)

        # return mesh
        tri_pts = np.array(tri_mesh.points)
        tri_elems = np.array(tri_mesh.elements)
        tri_e_atts = np.array(tri_mesh.element_attributes, dtype='int')

        tri_faces = np.array(tri_mesh.faces)
        tri_f_atts = np.array(tri_mesh.face_markers)
        f_mask = tri_f_atts > 0
        tri_f = tri_faces[f_mask]
        tri_fa = tri_f_atts[f_mask] - 1

        return cls(tri_pts, tri_elems, tri_e_atts, tri_f, tri_fa)

    # ----------------------------------------------------------------------- #
    # String and Representation Functions                                     #
    # ----------------------------------------------------------------------- #
    def __str__(self):
        nv = len(self.points)
        nd = len(self.points[0])
        pt_fmt = '\t'
        pt_fmt += ', '.join(['{pt[' + str(i) + ']: e}' for i in range(nd)])

        str_str = 'Mesh Points: ' + str(nv) + '\n'
        str_str += ''.join([pt_fmt.format(pt=p) + '\n' for p in self.points])

        str_str += 'Mesh Elements: ' + str(len(self.elements)) + '\n'
        str_str += '\n'.join(['\t' + str(tuple(e))[1:-1] for e in
                              self.elements])

        try:
            str_str += '\nElement Attributes: '
            str_str += str(len(self.element_attributes)) + '\n'
            str_str += '\n'.join(['\t' + str(a) for a in
                                  self.element_attributes])
        except TypeError:
            pass

        try:
            str_str += '\nFacets: ' + str(len(self.facets)) + '\n'
            str_str += '\n'.join(['\t' + str(tuple(f))[1:-1] for f in
                                  self.facets])
        except TypeError:
            pass

        try:
            str_str += '\nFacet Attributes: '
            str_str += str(len(self.facet_attributes)) + '\n'
            str_str += '\n'.join(['\t' + str(a) for a in
                                  self.facet_attributes])
        except TypeError:
            pass

        return str_str

    def __repr__(self):
        repr_str = 'TriMesh('
        repr_str += ', '.join([repr(v) for v in (self.points, self.elements,
                               self.element_attributes, self.facets,
                               self.facet_attributes)])
        repr_str += ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Write Function                                                          #
    # ----------------------------------------------------------------------- #
    def write(self, filename, format='txt', seeds=None, polymesh=None):
        """Write mesh to file.

        This function writes the contents of the mesh to a file.
        The format options are 'abaqus', 'tet/tri', 'txt', and 'vtk'.
        See the :ref:`s_tri_file_io` section of the :ref:`c_file_formats`
        guide for more details on these formats.

        Args:
            filename (str): The name of the file to write. In the cases of
                TetGen/Triangle, this is the basename of the files.
            format (str): {'abaqus' | 'tet/tri' | 'txt' | 'vtk'}
                *(optional)* The format of the output file.
                Default is 'txt'.
            seeds (SeedList): *(optional)* List of seeds. If given, VTK files
                will also include the phase number of of each element in the
                mesh. This assumes the ``element_attributes``
                field contains the seed number of each element.
            polymesh (PolyMesh): *(optional)* Polygonal mesh used for
                generating the triangular mesh. If given, will add surface
                unions to Abaqus files - for easier specification of
                boundary conditions.

        """  # NOQA: E501
        fmt = format.lower()
        if fmt == 'abaqus':
            # write top matter
            abaqus = '*Heading\n'
            abaqus += '** Job name: microstructure '
            abaqus += 'Model name: microstructure_model\n'
            abaqus += '** Generated by: MicroStructPy\n'

            # write parts
            abaqus += '**\n** PARTS\n**\n'
            abaqus += '*Part, name=Part-1\n'

            abaqus += '*Node\n'
            abaqus += ''.join([str(i + 1) + ''.join([', ' + str(x) for x in
                               pt]) + '\n' for i, pt in
                               enumerate(self.points)])

            n_dim = len(self.points[0])
            elem_type = {2: 'CPS3', 3: 'C3D4'}[n_dim]

            abaqus += '*Element, type=' + elem_type + '\n'
            abaqus += ''.join([str(i + 1) + ''.join([', ' + str(kp + 1) for kp
                                                     in elem]) + '\n' for
                               i, elem in enumerate(self.elements)])

            # Element sets - seed number
            elset_n_per = 16
            elem_atts = np.array(self.element_attributes)
            for att in np.unique(elem_atts):
                elset_name = 'Set-E-Seed-' + str(att)
                elset_str = '*Elset, elset=' + elset_name + '\n'
                elem_groups = [[]]
                for elem_ind, elem_att in enumerate(elem_atts):
                    if ~np.isclose(elem_att, att):
                        continue
                    if len(elem_groups[-1]) >= elset_n_per:
                        elem_groups.append([])
                    elem_groups[-1].append(elem_ind + 1)
                for group in elem_groups:
                    elset_str += ','.join([str(i) for i in group])
                    elset_str += '\n'

                abaqus += elset_str

            # Element Sets - phase number
            if seeds is not None:
                phase_nums = np.array([seed.phase for seed in seeds])
                for phase_num in np.unique(phase_nums):
                    mask = phase_nums == phase_num
                    seed_nums = np.nonzero(mask)[0]

                    elset_name = 'Set-E-Material-' + str(phase_num)
                    elset_str = '*Elset, elset=' + elset_name + '\n'
                    groups = [[]]
                    for seed_num in seed_nums:
                        if seed_num not in elem_atts:
                            continue
                        if len(groups[-1]) >= elset_n_per:
                            groups.append([])
                        seed_elset_name = 'Set-E-Seed-' + str(seed_num)
                        groups[-1].append(seed_elset_name)
                    for group in groups:
                        elset_str += ','.join(group)
                        elset_str += '\n'
                    abaqus += elset_str

            # Surfaces - Exterior and Interior
            facets = np.array(self.facets)
            facet_atts = np.array(self.facet_attributes)

            face_ids = {2: [2, 3, 1], 3: [3, 4, 2, 1]}[n_dim]

            for att in np.unique(facet_atts):
                facet_name = 'Surface-' + str(att)
                surf_str = '*Surface, name=' + facet_name + ', type=element\n'

                att_facets = facets[facet_atts == att]
                for facet in att_facets:
                    mask = np.isin(self.elements, facet)
                    n_match = mask.astype('int').sum(axis=1)
                    i_elem = np.argmax(n_match)
                    elem_id = i_elem + 1

                    i_missing = np.argmin(mask[i_elem])
                    face_id = face_ids[i_missing]

                    surf_str += str(elem_id) + ', S' + str(face_id) + '\n'

                abaqus += surf_str

            # Surfaces - Exterior
            poly_neighbors = np.array(polymesh.facet_neighbors)
            poly_mask = np.any(poly_neighbors < 0, axis=1)
            neigh_nums = np.min(poly_neighbors, axis=1)
            u_neighs = np.unique(neigh_nums[poly_mask])
            for neigh_num in u_neighs:
                mask = neigh_nums == neigh_num
                facet_name = 'Ext-Surface-' + str(-neigh_num)
                surf_str = '*Surface, name=' + facet_name + ', combine=union\n'
                for i, flag in enumerate(mask):
                    if flag:
                        surf_str += 'Surface-' + str(i) + '\n'
                abaqus += surf_str

            # End Part
            abaqus += '*End Part\n\n'

            # Assembly
            abaqus += '**\n'
            abaqus += '** ASSEMBLY\n'
            abaqus += '**\n'

            abaqus += '*Assembly, name=assembly\n'
            abaqus += '**\n'

            # Instances
            abaqus += '*Instance, name=I-Part-1, part=Part-1\n'
            abaqus += '*End Instance\n'

            # End Assembly
            abaqus += '**\n'
            abaqus += '*End Assembly\n'

            with open(filename, 'w') as file:
                file.write(abaqus)
        elif fmt in ('str', 'txt'):
            with open(filename, 'w') as file:
                file.write(str(self) + '\n')

        elif fmt == 'tet/tri':
            # create boundary markers
            bnd_mkrs = np.full(len(self.points), 0, dtype='int')

            facet_arr = np.array(self.facets)
            f_bnd_mkrs = np.full(len(self.facets), 0, dtype='int')
            elem_arr = np.array(self.elements)
            for elem in self.elements:
                for i in range(len(elem)):
                    e_facet = np.delete(elem, i)
                    f_mask = np.full(elem_arr.shape[0], True)
                    for kp in e_facet:
                        f_mask &= np.any(elem_arr == kp, axis=-1)

                    if np.sum(f_mask) == 1:
                        bnd_mkrs[e_facet] = 1

                        f_mask = np.full(facet_arr.shape[0], True)
                        for kp in e_facet:
                            f_mask &= np.any(facet_arr == kp, axis=-1)
                        f_bnd_mkrs[f_mask] = 1

            # write vertices
            n_pts, n_dim = np.array(self.points).shape
            nodes = ' '.join([str(n) for n in (n_pts, n_dim, 0, 1)]) + '\n'
            nodes += ''.join([str(i) + ''.join([' ' + str(x) for x in pt]) +
                              ' ' + str(bnd_mkrs[i]) + '\n' for i, pt in
                              enumerate(self.points)])

            with open(filename + '.node', 'w') as file:
                file.write(nodes)

            # write elements
            n_ele, n_kp = np.array(self.elements).shape
            is_att = self.element_attributes is not None
            n_att = int(is_att)
            eles = ' '.join([str(n) for n in (n_ele, n_kp, n_att)]) + '\n'
            for i, simplex in enumerate(self.elements):
                e_str = ' '.join([str(kp) for kp in simplex])
                if is_att:
                    e_str += ' ' + str(self.element_attributes[i])
                e_str += '\n'
                eles += e_str

            with open(filename + '.ele', 'w') as file:
                file.write(eles)

            # Write edges/faces
            if self.facets is not None:
                ext = {2: '.edge', 3: '.face'}[n_dim]

                n_facet, n_kp = np.array(self.facets).shape
                edge = ' '.join([str(n) for n in (n_facet, n_kp, 1)])
                edge += ''.join([str(i) + ''.join([' ' + str(k) for k in f]) +
                                 ' ' + str(mkr) + '\n' for f, mkr in
                                 zip(self.facets, f_bnd_mkrs)])
                with open(filename + ext, 'w') as file:
                    file.write(edge)

        elif fmt == 'vtk':
            n_kp = len(self.elements[0])
            mesh_type = {3: 'Triangular', 4: 'Tetrahedral'}[n_kp]
            pt_fmt = '{: f} {: f} {: f}\n'
            # write heading
            vtk = '# vtk DataFile Version 2.0\n'
            vtk += '{} mesh\n'.format(mesh_type)
            vtk += 'ASCII\n'
            vtk += 'DATASET UNSTRUCTURED_GRID\n'

            # Write points
            vtk += 'POINTS ' + str(len(self.points)) + ' float\n'
            if len(self.points[0]) == 2:
                vtk += ''.join([pt_fmt.format(x, y, 0) for x, y in
                                self.points])
            else:
                vtk += ''.join([pt_fmt.format(x, y, z) for x, y, z in
                                self.points])

            # write elements
            n_elem = len(self.elements)
            cell_fmt = str(n_kp) + n_kp * ' {}' + '\n'
            cell_sz = (1 + n_kp) * n_elem
            vtk += '\nCELLS ' + str(n_elem) + ' ' + str(cell_sz) + '\n'
            vtk += ''.join([cell_fmt.format(*el) for el in self.elements])

            # write cell type
            vtk += '\nCELL_TYPES ' + str(n_elem) + '\n'
            cell_type =  {3: '5', 4: '10'}[n_kp]
            vtk += ''.join(n_elem * [cell_type + '\n'])

            # write element attributes
            try:
                int(self.element_attributes[0])
                att_type = 'int'
            except TypeError:
                att_type = 'float'

            vtk += '\nCELL_DATA ' + str(n_elem) + '\n'
            vtk += 'SCALARS element_attributes ' + att_type + ' 1 \n'
            vtk += 'LOOKUP_TABLE element_attributes\n'
            vtk += ''.join([str(a) + '\n' for a in self.element_attributes])

            # Write phase numbers
            if seeds is not None:
                vtk += '\nSCALARS phase_numbers int 1 \n'
                vtk += 'LOOKUP_TABLE phase_numbers\n'
                vtk += ''.join([str(seeds[a].phase) + '\n' for a in
                                self.element_attributes])

            with open(filename, 'w') as file:
                file.write(vtk)

        else:
            e_str = 'Cannot write file type ' + str(format) + ' yet.'
            raise NotImplementedError(e_str)

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, index_by='element', material=[], loc=0, **kwargs):
        """Plot the mesh.

        This method plots the mesh using matplotlib.
        In 2D, this creates a :class:`matplotlib.collections.PolyCollection`
        and adds it to the current axes.
        In 3D, it creates a
        :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection` and
        adds it to the current axes.
        The keyword arguments are passed though to matplotlib.

        Args:
            index_by (str): *(optional)* {'element' | 'attribute'}
                Flag for indexing into the other arrays passed into the
                function. For example,
                ``plot(index_by='attribute', color=['blue', 'red'])`` will plot
                the elements with ``element_attribute`` equal to 0 in blue, and
                elements with ``element_attribute`` equal to 1 in red.
                Note that in 3D the facets are plotted instead of the elements,
                so kwarg lists must be based on ``facets`` and
                ``facet_attributes``. Defaults to 'element'.
            material (list): *(optional)* Names of material phases. One entry
                per material phase (the ``index_by`` argument is ignored).
                If this argument is set, a legend is added to the plot with
                one entry per material. Note that the ``element_attributes``
                in 2D or the ``facet_attributes`` in 3D must be the material
                numbers for the legend to be formatted properly.
            loc (int or str): *(optional)* The location of the legend,
                if 'material' is specified. This argument is passed directly
                through to :func:`matplotlib.pyplot.legend`. Defaults to 0,
                which is 'best' in matplotlib.
            **kwargs: Keyword arguments that are passed through to matplotlib.

        """
        n_dim = len(self.points[0])
        if n_dim == 2:
            ax = plt.gca()
        else:
            ax = plt.gcf().gca(projection=Axes3D.name)
        n_obj = _misc.ax_objects(ax)
        if n_obj > 0:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        else:
            xlim = [float('inf'), -float('inf')]
            ylim = [float('inf'), -float('inf')]
        if n_dim == 2:
            simps = np.array(self.elements)
            pts = np.array(self.points)
            xy = pts[simps, :]

            plt_kwargs = {}
            for key, value in kwargs.items():
                if type(value) in (list, np.array):
                    plt_value = []
                    for e_num, e_att in enumerate(self.element_attributes):
                        if index_by == 'element':
                            ind = e_num
                        elif index_by == 'attribute':
                            ind = int(e_att)
                        else:
                            e_str = 'Cannot index by {}.'.format(index_by)
                            raise ValueError(e_str)
                        v = value[ind]
                        plt_value.append(v)
                else:
                    plt_value = value
                plt_kwargs[key] = plt_value

            pc = collections.PolyCollection(xy, **plt_kwargs)
            ax.add_collection(pc)
            ax.autoscale_view()
        else:
            if n_obj > 0:
                zlim = ax.get_zlim()
            else:
                zlim = [float('inf'), -float('inf')]

            xy = [np.array([self.points[kp] for kp in f]) for f in self.facets]

            plt_kwargs = {}
            for key, value in kwargs.items():
                if type(value) in (list, np.array):
                    plt_value = []
                    for f_num, f_att in enumerate(self.facet_attributes):
                        if index_by == 'element':
                            ind = f_num
                        elif index_by == 'attribute':
                            ind = int(f_att)
                        else:
                            e_str = 'Cannot index by {}.'.format(index_by)
                            raise ValueError(e_str)
                        if ind < len(value):
                            v = value[ind]
                        else:
                            v = 'none'
                        plt_value.append(v)
                else:
                    plt_value = value
                plt_kwargs[key] = plt_value
            pc = Poly3DCollection(xy, **plt_kwargs)
            ax.add_collection(pc)

        # Add legend
        if material and index_by == 'attribute':
            p_kwargs = [{'label': m} for m in material]
            for key, value in kwargs.items():
                if type(value) not in (list, np.array):
                    for kws in p_kwargs:
                        kws[key] = value

                for i, m in enumerate(material):
                    if type(value) in (list, np.array):
                        p_kwargs[i][key] = value[i]
                    else:
                        p_kwargs[i][key] = value

            # Replace plural keywords
            for p_kw in p_kwargs:
                for kw in _misc.mpl_plural_kwargs:
                    if kw in p_kw:
                        p_kw[kw[:-1]] = p_kw[kw]
                        del p_kw[kw]
            handles = [patches.Patch(**p_kw) for p_kw in p_kwargs]
            ax.legend(handles=handles, loc=loc)

        # Adjust Axes
        mins = np.array(self.points).min(axis=0)
        maxs = np.array(self.points).max(axis=0)
        xlim = (min(xlim[0], mins[0]), max(xlim[1], maxs[0]))
        ylim = (min(ylim[0], mins[1]), max(ylim[1], maxs[1]))
        if n_dim == 2:
            plt.axis('square')
            plt.xlim(xlim)
            plt.ylim(ylim)
        elif n_dim == 3:
            zlim = (min(zlim[0], mins[2]), max(zlim[1], maxs[2]))
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_zlim(zlim)
            _misc.axisEqual3D(ax)


def facet_check(neighs, polymesh, phases):
    if any([n < 0 for n in neighs]):
        add_facet = True
    else:
        seed_nums = [polymesh.seed_numbers[n] for n in neighs]
        phase_nums = [polymesh.phase_numbers[n] for n in neighs]
        m1, m2 = [phases[n].get('material_type', 'solid') for n in
                  phase_nums]

        same_seed = seed_nums[0] == seed_nums[1]
        same_phase = phase_nums[0] == phase_nums[1]

        if (m1 in _misc.kw_solid) and same_seed:
            add_facet = False
        elif (m1 in _misc.kw_amorph) and same_phase:
            add_facet = False
        elif (m1 in _misc.kw_void) and (m2 in _misc.kw_void):
            add_facet = False
        else:
            add_facet = True

    return add_facet



"""N-Dimensional Box

This module contains the NBox class.

"""
# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #

from __future__ import division

import numpy as np

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# NBox Class                                                                  #
#                                                                             #
# --------------------------------------------------------------------------- #
class NBox(object):
    """N-dimensional box

    This class contains a generic, n-dimensinoal box.

    Args:
        side_lengths (list): *(optional)* Side lengths.
        center (list): *(optional)* Center of box.
        corner (list): *(optional)* Bottom-left corner.
        bounds (list): *(optional)* Bounds of box. Expected in the form
            [(xmin, xmax), (ymin, ymax), ...].
        limits : Alias for *bounds*.
        matrix (list, numpy.ndarray): *(optional)* Rotation matrix, nxn
    """

    def __init__(self, **kwargs):
        if 'bounds' in kwargs:
            lims = kwargs['bounds']
        elif 'limits' in kwargs:
            lims = kwargs['limits']

        if ('bounds' in kwargs) or ('limits' in kwargs):
            self.center = [0.5 * (lb + ub) for lb, ub in lims]
            self.side_lengths = [ub - lb for lb, ub in lims]

        elif ('side_lengths' in kwargs) and ('center' in kwargs):
            self.center = kwargs['center']
            self.side_lengths = kwargs['side_lengths']

        elif ('side_lengths' in kwargs) and ('corner' in kwargs):
            corner = kwargs['corner']
            side_lens = kwargs['side_lengths']
            cen = [xc + 0.5 * sl for xc, sl in zip(corner, side_lens)]
            self.center = cen
            self.side_lengths = side_lens

        elif ('center' in kwargs) and ('corner' in kwargs):
            corner = kwargs['corner']
            center = kwargs['center']
            side_lens = [2 * abs(x1 - x2) for x1, x2 in zip(center, corner)]
            self.center = center
            self.side_lengths = side_lens

        elif 'center' in kwargs:
            self.center = kwargs['center']
            self.side_lengths = [1 for _ in self.center]

        elif 'side_lengths' in kwargs:
            self.side_lengths = kwargs['side_lengths']
            self.center = [0 for _ in self.side_lengths]

        elif 'corner' in kwargs:
            self.side_lengths = [1 for _ in kwargs['corner']]
            self.center = [xc + 0.5 for xc in kwargs['corner']]

        if 'matrix' in kwargs:
            self.matrix = kwargs['matrix']
        else:
            try:
                self.matrix = np.eye(self.n_dim)
            except AttributeError:
                self.matrix = np.eye(len(self.side_lengths))

    # ----------------------------------------------------------------------- #
    # String and Representation Functions                                     #
    # ----------------------------------------------------------------------- #
    def __str__(self):
        cen = np.array(self.center)
        sides = np.array(self.side_lengths)
        cen_str = np.array2string(cen, separator=', ')
        sides_str = np.array2string(sides, separator=', ')

        str_str = 'Center: ' + cen_str + '\n'
        str_str += 'Side Lengths: ' + sides_str + '\n'
        str_str += 'Matrix: ('
        for row in self.matrix:
            str_str += '('
            str_str += ', '.join([str(val) for val in row])
            str_str += '),'
        str_str = str_str[:-1] + ')'
        return str_str

    def __repr__(self):
        repr_str = 'NBox('
        repr_str += 'center=' + repr(tuple(self.center)) + ', '
        repr_str += 'side_lengths=' + repr(tuple(self.side_lengths)) + ', '
        repr_str += 'matrix='
        repr_str += repr(tuple([tuple(r) for r in self.matrix])) + ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Dimension Getters                                                       #
    # ----------------------------------------------------------------------- #
    @property
    def corner(self):
        """list: bottom-left corner"""
        c_rel = -0.5 * np.array(self.matrix).dot(self.side_lengths)
        return np.array(self.center) + c_rel

    @property
    def limits(self):
        """list: (lower, upper) bounds of the box"""
        cen = self.center
        side_lens = self.side_lengths

        n_dim = len(cen)
        powers = np.power(2, np.arange(n_dim))
        n_pts = np.power(2, n_dim)
        pts = np.zeros((n_pts, n_dim))
        for i in range(n_pts):
            for j in range(n_dim):
                ind = (i // powers[j]) % 2
                pts[i, j] = (ind - 0.5) * side_lens[j]

        pts = np.array(cen) + pts.dot(np.array(self.matrix).T)
        mins = pts.min(axis=0)
        maxs = pts.max(axis=0)
        return np.array([mins, maxs]).T

    @property
    def bounds(self):
        """list: (lower, upper) bounds of the box"""
        return self.limits

    # ----------------------------------------------------------------------- #
    # Sample Limits                                                           #
    # ----------------------------------------------------------------------- #
    @property
    def sample_limits(self):
        """ list: (lower, upper) bounds of the sampling region of the box"""
        cen = self.center
        lens = self.side_lengths
        tol = 1e-4 * max(lens)
        return [(x - 0.5 * s + tol, x + 0.5 * s - tol) for x, s in
                zip(cen, lens)]

    # ----------------------------------------------------------------------- #
    # Area/Volume                                                             #
    # ----------------------------------------------------------------------- #
    @property
    def n_vol(self):
        """float: area, volume of n-box"""
        return np.prod(self.side_lengths)

    # ----------------------------------------------------------------------- #
    # Within Test                                                             #
    # ----------------------------------------------------------------------- #
    def within(self, points):
        """Test if points are within n-box.

        This function tests whether a point or set of points are within the
        n-box. For the set of points, a list of booleans is returned to
        indicate which points are within the n-box.

        Args:
            points (list or numpy.ndarray): Point or list of points.

        Returns:
            bool or numpy.ndarray: Flags set to True for points in geometry.
        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        min_dist = 0.5 * np.array(self.side_lengths)

        mask = np.all(np.abs(rel_pos) <= min_dist, axis=-1)
        if single_pt:
            return mask[0]
        else:
            return mask



from __future__ import division

import os

import numpy as np
import scipy.integrate
import scipy.stats
from matplotlib import pyplot as plt

import microstructpy as msp

# Define the domain
domain = msp.geometry.Square(corner=(0, 0), side_length=10)

# Define the material phases
a_dist = scipy.stats.lognorm(s=1, scale=0.1)
matrix_phase = {'fraction': 1,
                'material_type': 'matrix',
                'shape': 'circle',
                'area': a_dist}

neighborhood_phase = {'fraction': 1,
                      'material_type': 'solid',
                      'shape': 'ellipse',
                      'a': 1.5,
                      'b': 0.6,
                      'angle_deg': scipy.stats.uniform(0, 360)}

phases = [matrix_phase, neighborhood_phase]

# Create the seed list
seeds = msp.seeding.SeedList.from_info(phases, domain.area)
seeds.position(domain)

# Replace the neighborhood phase with materials
a = neighborhood_phase['a']
b = neighborhood_phase['b']
r = b / 3
n = 16

t_perim = np.linspace(0, 2 * np.pi, 201)
x_perim = (a - r) * np.cos(t_perim)
y_perim = (b - r) * np.sin(t_perim)
dx = np.insert(np.diff(x_perim), 0, 0)
dy = np.insert(np.diff(y_perim), 0, 0)
ds = np.sqrt(dx * dx + dy * dy)
arc_len = scipy.integrate.cumtrapz(ds, x=t_perim, initial=0)
eq_spaced = arc_len[-1] * np.arange(n) / n
x_pts = np.interp(eq_spaced, arc_len, x_perim)
y_pts = np.interp(eq_spaced, arc_len, y_perim)

repl_seeds = msp.seeding.SeedList()
geom = {'a': a - 2 * r, 'b': b - 2 * r}
for sn, seed in enumerate(seeds):
    if seed.phase == 0:
        repl_seeds.append(seed)
    else:
        center = seed.position
        theta = seed.geometry.angle_rad

        geom['angle_rad'] = theta
        geom['center'] = center
        core_seed = msp.seeding.Seed.factory('ellipse', phase=3,
                                             position=seed.position, **geom)
        repl_seeds.append(core_seed)

        x_ring = center[0] + x_pts * np.cos(theta) - y_pts * np.sin(theta)
        y_ring = center[1] + x_pts * np.sin(theta) + y_pts * np.cos(theta)
        for i in range(n):
            phase = 1 + (i % 2)
            center = (x_ring[i], y_ring[i])
            ring_geom = {'center': center, 'r': r}
            ring_seed = msp.seeding.Seed.factory('circle', position=center,
                                                 phase=phase, **ring_geom)
            if domain.within(center):
                repl_seeds.append(ring_seed)

# Create polygon and triangle meshes
pmesh = msp.meshing.PolyMesh.from_seeds(repl_seeds, domain)
phases = [{'material_type': 'solid'} for i in range(4)]
phases[0]['material_type'] = 'matrix'
tmesh = msp.meshing.TriMesh.from_polymesh(pmesh, phases, min_angle=20,
                                          max_volume=0.1)

# Plot triangle mesh
colors = ['C' + str(repl_seeds[att].phase) for att in tmesh.element_attributes]
tmesh.plot(facecolors=colors, edgecolors='k', linewidth=0.2)

plt.axis('square')
plt.xlim(domain.limits[0])
plt.ylim(domain.limits[1])

file_dir = os.path.dirname(os.path.realpath(__file__))
filename = os.path.join(file_dir, 'grain_neighborhoods/trimesh.png')
dirs = os.path.dirname(filename)
if not os.path.exists(dirs):
    os.makedirs(dirs)
plt.savefig(filename, bbox_inches='tight', pad_inches=0)



from __future__ import division

import os
import sys

import numpy as np

from microstructpy import geometry
from microstructpy.meshing.polymesh import PolyMesh
from microstructpy.meshing.polymesh import kp_loop
from microstructpy.seeding import Seed
from microstructpy.seeding import SeedList

'''
Test Mesh

    D ------- C
    |         | \
    |         |  \
    |         |   \
    A ------- B -- E

    A = (0, 0)
    B = (1, 0)
    C = (1, 1)
    D = (0, 1)
    E = (1.5, 0)

    Facets:
        0 A B
        1 B C
        2 C D
        3 D A
        4 B E
        5 E C

    Regions:
        0, 1, 2, 3
        4, 5, 1

    Seed Numbers:
        0
        1

    Phase Numbers:
        2
        2

'''

A = (0, 0)
B = (1, 0)
C = (1, 1)
D = (0, 1)
E = (1.5, 0)
pts = [A, B, C, D, E]

facets = [(0, 1), (1, 2), (2, 3), (3, 0), (1, 4), (4, 2)]

regions = [(0, 1, 2, 3), (4, 5, 1)]

seed_nums = [0, 1]
phase_nums = [2, 2]

pmesh = PolyMesh(pts, facets, regions, seed_nums, phase_nums)


def test_eq():
    assert pmesh == pmesh

    r2 = [(0, 3, 2, 1), (4, 1, 5)]
    pmesh2 = PolyMesh(pts, facets, r2, seed_nums, phase_nums)
    assert pmesh == pmesh2
    assert pmesh2 == pmesh

    assert pmesh != pts
    assert pmesh != PolyMesh(pts, facets, regions, seed_nums, phase_nums[:-1])

    r3 = [(0, 3, 2), (4, 1, 5)]
    assert pmesh != PolyMesh(pts, facets, r3, seed_nums, phase_nums)

    pt2 = [(-1, 0), B, C, D, E]
    assert pmesh != PolyMesh(pt2, facets, regions, seed_nums, phase_nums)

    f2 = [(0, 1), (1, 3), (2, 3), (3, 0), (1, 4), (4, 2)]
    assert pmesh != PolyMesh(pts, f2, regions, seed_nums, phase_nums)

    pmesh3 = PolyMesh(pts, facets, regions)
    assert pmesh3 == pmesh3
    assert pmesh3 != pmesh


def test_read_write():
    pyver_str = sys.version.split(' ')[0]
    pyver_str = pyver_str.replace('.', '_')

    fname = 'tmp_' + pyver_str + '.txt'
    filepath = os.path.dirname(__file__)
    filename = filepath + '/' + fname

    pmesh.write(filename)
    rw_pmesh = PolyMesh.from_file(filename)
    os.remove(filename)
    assert pmesh == rw_pmesh

    alt_str = 'comments\n' + str(pmesh)
    with open(filename, 'w') as file:
        file.write(alt_str + '\n')
    rw_pmesh = PolyMesh.from_file(filename)
    os.remove(filename)
    assert pmesh == rw_pmesh


def test_repr():
    assert pmesh == eval(repr(pmesh))

    pmesh2 = PolyMesh(pts, facets, regions)
    assert pmesh2 == eval(repr(pmesh2))


def test_from_seeds():
    # d^2 = dx^2 + dy^2 - r^2
    # Let d = 3
    # And circle 1 has radius 4,
    # so dx = 5
    # And circle 2 has radius 2,
    # so dx = np.sqrt(13)
    # and in both cases, y= 0

    d = 3
    r1 = 4
    r2 = 2

    x1 = -np.sqrt(d * d + r1 * r1)
    x2 = np.sqrt(d * d + r2 * r2)

    p1 = 2
    p2 = 3

    x_len = 2 * 1.1 * max(-x1 + r1, x2 + r2)
    y_len = 2 * 1.1 * max(r1, r2)

    # create polymesh by hand
    A = (-0.5 * x_len, -0.5 * y_len)
    B = (0, -0.5 * y_len)
    C = (0.5 * x_len, -0.5 * y_len)
    D = (0.5 * x_len, 0.5 * y_len)
    E = (0, 0.5 * y_len)
    F = (-0.5 * x_len, 0.5 * y_len)

    pts = [A, B, C, D, E, F]
    facets = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0), (1, 4)]
    regions = [(0, 6, 4, 5), (1, 2, 3, 6)]
    poly_exp = PolyMesh(pts, facets, regions, [0, 1], [p1, p2])

    # create polymesh from seeds and domain
    s1 = Seed.factory('circle', r=r1, center=(x1, 0), phase=p1)
    s2 = Seed.factory('circle', r=r2, center=(x2, 0), phase=p2)
    slist = SeedList([s1, s2])

    lens = [x_len, y_len]
    dom = geometry.Rectangle(center=(0.0, 0.0), side_lengths=lens)

    poly_act = PolyMesh.from_seeds(slist, dom)
    assert poly_exp == poly_act


def test_kp_loop():
    p1 = [0, 1]
    p2 = [1, 2]
    p3 = [2, 0]
    pairs = [p1, p2, p3]
    loop = kp_loop(pairs)

    possible_pairs = [(0, 1, 2),
                      (1, 2, 0),
                      (2, 0, 1),
                      (2, 1, 0),
                      (1, 0, 2),
                      (0, 2, 1)]
    assert tuple(loop) in possible_pairs



"""Seed List

This module contains the class definition for the SeedList class.
"""
# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #

from __future__ import division
from __future__ import print_function

import warnings

import aabbtree
import numpy as np
import scipy.stats
from matplotlib import collections
from matplotlib import patches
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from mpl_toolkits.mplot3d import Axes3D
from pyquaternion import Quaternion
from scipy.spatial import distance

from microstructpy import _misc
from microstructpy import geometry
from microstructpy.seeding import seed as _seed

__all__ = ['SeedList']
__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# SeedList Class                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
class SeedList(object):
    """List of seed geometries.

    The SeedList is similar to a standard Python list, but contains instances
    of the :class:`.Seed` class. It can be generated from a list of Seeds,
    by creating enough seeds to fill a given volume, or by reading the content
    of a cache text file.

    Args:
        seeds (list): *(optional)* List of :class:`.Seed` instances.

    """
    # ----------------------------------------------------------------------- #
    # Constructors                                                            #
    # ----------------------------------------------------------------------- #
    def __init__(self, seeds=[]):
        self.seeds = seeds

    @classmethod
    def from_file(cls, filename):
        """Create seed list from file containing list of seeds

        This function creates a seed list from a file containing a list of
        seeds. This file should contain the string representations of seeds,
        separated by a newline character (which is the behavior of
        :meth:`write`).

        Args:
            filename (str): File containing the seed list.

        Returns:
            SeedList: Instance of class.
        """
        with open(filename, 'r') as file:
            file_str = file.read()

        beg = 'Geometry:'
        rem = file_str.split(beg)[1:]

        return cls([_seed.Seed.from_str(beg + s) for s in rem])

    @classmethod
    def from_info(cls, phases, volume, rng_seeds={}):
        """Create seed list from microstructure information

        This function creates a seed list from information about the
        microstruture. The "phases" input should be a list of material
        phase dictionaries, formatted according to the :ref:`phase_dict_guide`
        guide.

        The "volume" input is the minimum volume of the list of seeds. Seeds
        will be added to the list until this volume threshold is crossed.

        Finally, the "rng_seeds" input is a dictionary of random number
        generator (RNG) seeds for each parameter of the seed geometries.
        For example, if one of the phases uses "size" to define the seeds,
        then "size" could be a keyword of the "rng_seeds" input. The value
        should be a non-negative integer, to seed the RNG for size.
        The default RNG seed is 0.

        Note:
            If two or more parameters have the same RNG seed and the same
            kernel of the distribution, those parameters will **not** be
            correlated. This method updates RNG seeds based on the order that
            distributions are sampled to avoid correlation between independent
            random variables.

        Args:
            phases (dict): Dictionary of phase information, see
                :ref:`phase_dict_guide` for a guide.
            volume (float): The total area/volume of the seeds in the list.
            rng_seeds (dict): *(optional)* Dictionary of RNG seeds for each
                step in the seeding process. The dictionary keys should match
                shape parameters in ``phases``. For example::

                    rng_seeds = {
                        'size': 0,
                        'angle': 3,
                    }

        Returns:
            SeedList: An instance of the class containing seeds prescribed by
            the phase information and filling the given volume.

        """

        # determine dimensionality, set default shape
        default_shapes = {2: 'circle', 3: 'sphere'}
        n_dim = None
        if isinstance(phases, dict):
            phases = [phases]

        for phase in phases:
            if 'shape' in phase:
                n_dim = geometry.factory(phase['shape']).n_dim
        if n_dim is None:
            e_str = 'Number of dimensions could not be determined from phase '
            e_str += 'shapes. Consider setting the shape of a phase, or'
            e_str += ' specifying the number of dimensions.'
            raise ValueError(e_str)

        for phase in phases:
            if 'shape' in phase:
                assert geometry.factory(phase['shape']).n_dim == n_dim
            else:
                phase['shape'] = default_shapes[n_dim]

        # compute volume of each phase
        vol_rng = rng_seeds.get('fraction', 0)
        np.random.seed(vol_rng)

        n_phases = len(phases)
        rel_vols = np.ones(n_phases)
        for i, phase in enumerate(phases):
            vol = phase.get('fraction', 1)
            try:
                v_sample = -1
                while v_sample < 0:
                    v_sample = vol.rvs()
                rel_vols[i] = v_sample
            except AttributeError:
                rel_vols[i] = vol
        vol_fracs = rel_vols / sum(rel_vols)
        phase_vols = volume * vol_fracs

        # compute number of seeds for each phase
        if n_dim == 2:
            avg_vols = [geometry.factory(p['shape']).area_expectation(**p)
                        for p in phases]
        else:
            avg_vols = [geometry.factory(p['shape']).volume_expectation(**p)
                        for p in phases]
        weights = phase_vols / np.array(avg_vols)
        pop_fracs = weights / sum(weights)

        seed_vol = 0
        seeds = []
        max_int = np.iinfo(np.int32).max
        while seed_vol < volume:
            # Pick the phase
            rng_seed = rng_seeds.get('phase', 0)
            np.random.seed(rng_seed)
            phase_num = np.random.choice(n_phases, p=pop_fracs)
            phase = phases[phase_num]
            rng_seeds['phase'] = np.random.randint(max_int)

            # Create the seed
            seed_shape = phase['shape']
            seed_args = {'phase': phase_num}
            kw_n = 0
            for kw in set(phase) - set(_misc.gen_kws):
                # set the RNG seed
                rng_seed = rng_seeds.get(kw, 0)
                np.random.seed(rng_seed)

                # Sample, with special cases for orientation
                if kw not in _misc.ori_kws:
                    try:
                        val = phase[kw].rvs(random_state=rng_seed)
                    except AttributeError:
                        val = phase[kw]
                    seed_args[kw] = val
                elif (phase[kw] == 'random') and (n_dim == 2):
                    np.random.seed(rng_seed)
                    seed_args['angle_deg'] = 360 * np.random.rand()
                elif phase[kw] == 'random':
                    np.random.seed(rng_seed)
                    elems = np.random.normal(size=4)
                    mag = np.linalg.norm(elems)
                    elems /= mag
                    val = Quaternion(elems).rotation_matrix
                    seed_args[kw] = val
                elif kw in ['rot_seq', 'rot_seq_deg', 'rot_seq_rad']:
                    seq = []
                    val = phase[kw]
                    if not isinstance(val, list):
                        val = [val]
                    for rotation in val:
                        rot_dict = {str(kw): rotation[kw] for kw in rotation}
                        ax = rot_dict.get('axis', 'x')
                        ang_dist = rot_dict.get('angle', 0)
                        try:
                            ang = ang_dist.rvs(random_state=rng_seed)
                        except AttributeError:
                            ang = ang_dist
                        seq.append((ax, ang))
                    seed_args[kw] = seq
                else:
                    try:
                        val = phase[kw].rvs(random_state=rng_seed)
                    except AttributeError:
                        val = phase[kw]
                    seed_args[kw] = val

                # Update the RNG seed
                np.random.seed(rng_seed + kw_n)
                rng_seeds[kw] = np.random.randint(max_int - kw_n)
                kw_n += 1

            # Add seed to list
            seed = _seed.Seed.factory(seed_shape, **seed_args)
            seeds.append(seed)
            seed_vol += seed.volume

        return cls(seeds)

    # ----------------------------------------------------------------------- #
    # Representation and String Functions                                     #
    # ----------------------------------------------------------------------- #
    def __repr__(self):
        repr_str = 'SeedList('
        repr_str += repr(self.seeds)
        repr_str += ')'
        return repr_str

    def __str__(self):
        str_str = '\n'.join([str(s) for s in self.seeds])
        return str_str

    # ----------------------------------------------------------------------- #
    # List Methods                                                            #
    # ----------------------------------------------------------------------- #
    def __getitem__(self, i):
        if isinstance(i, slice):
            return SeedList(self.seeds[i])

        try:
            return self.seeds[i]
        except TypeError:
            indices = np.arange(len(self))[i]
            return SeedList([self.seeds[ind] for ind in indices])

    def __setitem__(self, i, s):
        try:
            self.seeds[int(i)] = s
        except TypeError:
            n_added = 0
            for ind, i_val in enumerate(i):
                if str(i_val) == 'True':
                    self.seeds[ind] = s[n_added]
                    n_added += 1
                elif str(i_val) != 'False':
                    self.seeds[int(i_val)] = s[ind]

    def __add__(self, seedlist):
        """Add seed lists together

        This function overloads the + operator, similar to
        :meth:`list.__add__`.

        Args:
            seedlist (SeedList, list): List of seeds to add.

        Returns:
            SeedList: Instance that joins the two seed lists.

        .. versionadded:: 1.1
        """
        if type(self) == type(seedlist):
            return SeedList(self.seeds + seedlist.seeds)
        else:
            return SeedList(self.seeds + seedlist)

    def __len__(self):
        return len(self.seeds)

    def append(self, seed):
        """Append seed

        This function appends a seed to the list.

        Args:
            seed (Seed): The seed to append to the list

        """
        self.seeds.append(seed)

    def extend(self, seeds):
        """Extend seed list

        This function adds a list of seeds to the end of the seed list.

        Args:
            seeds (list or SeedList): List of seeds

        """
        if isinstance(seeds, SeedList):
            self.seeds.extend(seeds.seeds)
        else:
            self.seeds.extend(seeds)

    # ----------------------------------------------------------------------- #
    # Comparison Methods                                                      #
    # ----------------------------------------------------------------------- #
    def __eq__(self, other_list):
        same = len(self) == len(other_list)
        if same:
            for s1, s2 in zip(self, other_list):
                same &= s1 == s2
        return same

    # ----------------------------------------------------------------------- #
    # Write Function                                                          #
    # ----------------------------------------------------------------------- #
    def write(self, filename, format='txt'):
        """Write seed list to a text file

        This function writes out the seed list to a file. The format of the
        file can be either 'txt' or 'vtk'. The content of the 'txt' file
        is human-readable and can be read by the
        :func:`SeedList.from_file` method.
        The 'vtk' option creates a VTK legacy file with the grain geometries.

        For grains that are non-spherical, the spherical breakdown of the seed
        is output instead of the seed itself.

        Args:
            filename (str): File to write the seed list.
        """

        if format == 'txt':
            with open(filename, 'w') as file:
                file.write(str(self) + '\n')
        elif format == 'vtk':
            # Unpack breakdowns
            bkdwns = np.array([b for s in self for b in s.breakdown])
            n_pts, n_col = bkdwns.shape
            seed_nums = np.array([i for i, s in enumerate(self) for b in
                                  s.breakdown])
            centers = np.zeros((n_pts, 3))
            centers[:, :(n_col - 1)] = bkdwns[:, :-1]
            diameters = 2 * bkdwns[:, -1]

            pt_fmt = '{: f} {: f} {: f}\n'
            # write heading
            vtk = '# vtk DataFile Version 2.0\n'
            vtk += 'Tetrahedral mesh\n'
            vtk += 'ASCII\n'
            vtk += 'DATASET POLYDATA\n'

            # Write points
            vtk += 'POINTS ' + str(n_pts) + ' float\n'
            vtk += ''.join([pt_fmt.format(*c) for c in centers])

            # Write diameters
            vtk += '\nPOINT_DATA ' + str(n_pts) + '\n'
            vtk += 'SCALARS diameters double 1 \n'
            vtk += 'LOOKUP_TABLE Diameters\n'
            vtk += ''.join([str(d) + '\n' for d in diameters])

            # Write material numbers
            vtk += '\nSCALARS phase_numbers int 1 \n'
            vtk += 'LOOKUP_TABLE phase_numbers\n'
            vtk += ''.join([str(self[n].phase) + '\n' for n in seed_nums])

            # Write seed numbers
            vtk += '\nSCALARS seed_numbers int 1 \n'
            vtk += 'LOOKUP_TABLE seed_numbers \n'
            vtk += ''.join([str(n) + '\n' for n in seed_nums])

            with open(filename, 'w') as file:
                file.write(vtk)

        else:
            raise ValueError('Cannot write to format ' + str(format))

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, index_by='seed', material=[], loc=0, **kwargs):
        """Plot the seeds in the seed list.

        This function plots the seeds contained in the seed list.
        In 2D, the seeds are grouped into matplotlib collections to reduce
        the computational load. In 3D, matplotlib does not have patches, so
        each seed is rendered as its own surface.

        Additional keyword arguments can be specified and passed through to
        matplotlib. These arguments should be either single values
        (e.g. ``edgecolors='k'``), or lists of values that have the same
        length as the seed list.

        Args:
            index_by (str): *(optional)* {'material' | 'seed'}
                Flag for indexing into the other arrays passed into the
                function. For example,
                ``plot(index_by='material', color=['blue', 'red'])`` will plot
                the seeds with ``phase`` equal to 0 in blue, and seeds with
                ``phase`` equal to 1 in red. Defaults to 'seed'.
            material (list): *(optional)* Names of material phases. One entry
                per material phase (the ``index_by`` argument is ignored).
                If this argument is set, a legend is added to the plot with
                one entry per material.
            loc (int or str): *(optional)* The location of the legend,
                if 'material' is specified. This argument is passed directly
                through to :func:`matplotlib.pyplot.legend`. Defaults to 0,
                which is 'best' in matplotlib.
            **kwargs: Keyword arguments to pass to matplotlib

        """
        seed_args = [{} for seed in self]
        for seed_num, seed in enumerate(self):
            phase_num = seed.phase
            for key, val in kwargs.items():
                if type(val) in (list, np.array):
                    if index_by == 'seed' and len(val) > seed_num:
                        seed_args[seed_num][key] = val[seed_num]
                    elif index_by == 'material' and len(val) > phase_num:
                        seed_args[seed_num][key] = val[phase_num]
                else:
                    seed_args[seed_num][key] = val

        n = self[0].geometry.n_dim
        if n == 2:
            ax = plt.gca()
        else:
            ax = plt.gcf().gca(projection=Axes3D.name)
        n_obj = _misc.ax_objects(ax)
        if n_obj > 0:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        else:
            xlim = [float('inf'), -float('inf')]
            ylim = [float('inf'), -float('inf')]

        if n == 3:
            if n_obj > 0:
                zlim = ax.get_zlim()
            else:
                zlim = [float('inf'), -float('inf')]

            for seed, args in zip(self, seed_args):
                seed.plot(**args)

        elif n == 2:
            ellipse_data = {'w': [], 'h': [], 'a': [], 'xy': []}
            ec_kwargs = {}

            rect_data = {'xy': [], 'w': [], 'h': [], 'angle': []}
            rect_kwargs = {}

            pc_verts = []
            pc_kwargs = {}
            for seed, args in zip(self, seed_args):
                geom_name = type(seed.geometry).__name__.lower().strip()
                if geom_name == 'ellipse':
                    a, b = seed.geometry.axes
                    cen = np.array(seed.position)
                    t = seed.geometry.angle_deg

                    ellipse_data['w'].append(2 * a)
                    ellipse_data['h'].append(2 * b)
                    ellipse_data['a'].append(t)
                    ellipse_data['xy'].append(cen)

                    for key, val in args.items():
                        val_list = ec_kwargs.get(key, [])
                        val_list.append(val)
                        ec_kwargs[key] = val_list

                elif geom_name == 'circle':
                    diam = seed.geometry.diameter
                    cen = np.array(seed.position)

                    ellipse_data['w'].append(diam)
                    ellipse_data['h'].append(diam)
                    ellipse_data['a'].append(0)
                    ellipse_data['xy'].append(cen)

                    for key, val in args.items():
                        val_list = ec_kwargs.get(key, [])
                        val_list.append(val)
                        ec_kwargs[key] = val_list

                elif geom_name in ['rectangle', 'square']:
                    w, h = seed.geometry.side_lengths
                    corner = seed.geometry.corner
                    t = seed.geometry.angle_deg

                    rect_data['w'].append(w)
                    rect_data['h'].append(h)
                    rect_data['angle'].append(t)
                    rect_data['xy'].append(corner)

                    for key, val in args.items():
                        val_list = rect_kwargs.get(key, [])
                        val_list.append(val)
                        rect_kwargs[key] = val_list

                elif geom_name == 'curl':
                    xy = seed.geometry.plot_xy()
                    pc_verts.append(xy)
                    for key, val in args.items():
                        val_list = pc_kwargs.get(key, [])
                        val_list.append(val)
                        pc_kwargs[key] = val_list

                elif geom_name == 'nonetype':
                    pass

                else:
                    e_str = 'Cannot plot groups of ' + geom_name
                    e_str += ' yet.'
                    raise NotImplementedError(e_str)

            # abbreviate kwargs if all the same
            for key, val in ec_kwargs.items():
                v1 = val[0]
                same = True
                for v in val:
                    same &= v == v1
                if same:
                    ec_kwargs[key] = v1

            for key, val in pc_kwargs.items():
                v1 = val[0]
                same = True
                for v in val:
                    same &= v == v1
                if same:
                    pc_kwargs[key] = v1

            # Plot Circles and Ellipses
            ax = plt.gca()

            w = np.array(ellipse_data['w'])
            h = np.array(ellipse_data['h'])
            a = np.array(ellipse_data['a'])
            xy = np.array(ellipse_data['xy'])
            ec = collections.EllipseCollection(w, h, a, units='x', offsets=xy,
                                               transOffset=ax.transData,
                                               **ec_kwargs)
            ax.add_collection(ec)

            # Plot Rectangles
            rects = [Rectangle(xy=xyi, width=wi, height=hi, angle=ai) for
                     xyi, wi, hi, ai in zip(rect_data['xy'], rect_data['w'],
                     rect_data['h'], rect_data['angle'])]
            rc = collections.PatchCollection(rects, False, **rect_kwargs)
            ax.add_collection(rc)

            # Plot Polygons
            pc = collections.PolyCollection(pc_verts, **pc_kwargs)
            ax.add_collection(pc)

            ax.autoscale_view()

        # Add legend
        if material:
            p_kwargs = [{'label': m} for m in material]
            if index_by == 'seed':
                for seed_kwargs, seed in zip(seed_args,  self):
                    p = seed.phase
                    p_kwargs[p].update(seed_kwargs)
            else:
                for key, val in kwargs.items():
                    if type(val) in (list, np.array):
                        for i, elem in enumerate(val):
                            p_kwargs[i][key] = elem
                    else:
                        for i in range(len(p_kwargs)):
                            p_kwargs[i][key] = val

            # Replace plural keywords
            for p_kw in p_kwargs:
                for kw in _misc.mpl_plural_kwargs:
                    if kw in p_kw:
                        p_kw[kw[:-1]] = p_kw[kw]
                        del p_kw[kw]
            handles = [patches.Patch(**p_kw) for p_kw in p_kwargs]
            ax.legend(handles=handles, loc=loc)

        # Adjust Axes
        seed_lims = [np.array(s.geometry.limits).flatten() for s in self]
        mins = np.array(seed_lims)[:, 0::2].min(axis=0)
        maxs = np.array(seed_lims)[:, 1::2].max(axis=0)
        xlim = (min(xlim[0], mins[0]), max(xlim[1], maxs[0]))
        ylim = (min(ylim[0], mins[1]), max(ylim[1], maxs[1]))
        if n == 2:
            plt.axis('square')
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
        if n == 3:
            zlim = (min(zlim[0], mins[2]), max(zlim[1], maxs[2]))
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_zlim(zlim)
            _misc.axisEqual3D(ax)

    def plot_breakdown(self, index_by='seed', material=[], loc=0, **kwargs):
        """Plot the breakdowns of the seeds in seed list.

        This function plots the breakdowns of seeds contained in the seed list.
        In 2D, the breakdowns are grouped into matplotlib collections to reduce
        the computational load. In 3D, matplotlib does not have patches, so
        each breakdown is rendered as its own surface.

        Additional keyword arguments can be specified and passed through to
        matplotlib. These arguments should be either single values
        (e.g. ``edgecolors='k'``), or lists of values that have the same
        length as the seed list.

        Args:
            index_by (str): *(optional)* {'material' | 'seed'}
                Flag for indexing into the other arrays passed into the
                function. For example,
                ``plot(index_by='material', color=['blue', 'red'])`` will plot
                the seeds with ``phase`` equal to 0 in blue, and seeds with
                ``phase`` equal to 1 in red. Defaults to 'seed'.
            material (list): *(optional)* Names of material phases. One entry
                per material phase (the ``index_by`` argument is ignored).
                If this argument is set, a legend is added to the plot with
                one entry per material.
            loc (int or str): *(optional)* The location of the legend,
                if 'material' is specified. This argument is passed directly
                through to :func:`matplotlib.pyplot.legend`. Defaults to 0,
                which is 'best' in matplotlib.
            **kwargs: Keyword arguments to pass to matplotlib

        """
        seed_args = [{} for seed in self]
        for seed_num, seed in enumerate(self):
            phase_num = seed.phase
            for key, val in kwargs.items():
                if type(val) in (list, np.array):
                    if index_by == 'seed' and len(val) > seed_num:
                        seed_args[seed_num][key] = val[seed_num]
                    elif index_by == 'material' and len(val) > phase_num:
                        seed_args[seed_num][key] = val[phase_num]
                else:
                    seed_args[seed_num][key] = val

        n = self[0].geometry.n_dim
        if n == 2:
            ax = plt.gca()
        else:
            ax = plt.gcf().gca(projection=Axes3D.name)
        n_obj = _misc.ax_objects(ax)
        if n_obj > 0:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        else:
            xlim = [float('inf'), -float('inf')]
            ylim = [float('inf'), -float('inf')]

        if n == 3:
            if n_obj > 0:
                zlim = ax.get_zlim()
            else:
                zlim = [float('inf'), -float('inf')]

            for seed, args in zip(self, seed_args):
                seed.plot_breakdown(**args)

        elif n == 2:
            breakdowns = np.zeros((0, 3))
            ec_kwargs = {}
            for seed, args in zip(self, seed_args):
                breakdowns = np.concatenate((breakdowns, seed.breakdown))
                n_c = len(seed.breakdown)
                for key, val in args.items():
                    val_list = ec_kwargs.get(key, [])
                    val_list.extend(n_c * [val])
                    ec_kwargs[key] = val_list
            d = 2 * breakdowns[:, -1]
            xy = breakdowns[:, :-1]
            a = np.full(len(breakdowns), 0)

            # abbreviate kwargs if all the same
            for key, val in ec_kwargs.items():
                v1 = val[0]
                same = True
                for v in val:
                    same &= v == v1
                if same:
                    ec_kwargs[key] = v1

            ax = plt.gca()
            ec = collections.EllipseCollection(d, d, a, units='x', offsets=xy,
                                               transOffset=ax.transData,
                                               **ec_kwargs)
            ax.add_collection(ec)
            ax.autoscale_view()

        # Add legend
        if material:
            p_kwargs = [{'label': m} for m in material]
            if index_by == 'seed':
                for seed_kwargs, seed in zip(seed_args,  self):
                    p = seed.phase
                    p_kwargs[p].update(seed_kwargs)
            else:
                for key, val in kwargs.items():
                    if type(val) in (list, np.array):
                        for i, elem in enumerate(val):
                            p_kwargs[i][key] = elem
                    else:
                        for i in range(len(p_kwargs)):
                            p_kwargs[i][key] = val

            # Replace plural keywords
            for p_kw in p_kwargs:
                for kw in _misc.mpl_plural_kwargs:
                    if kw in p_kw:
                        p_kw[kw[:-1]] = p_kw[kw]
                        del p_kw[kw]
            handles = [patches.Patch(**p_kw) for p_kw in p_kwargs]
            if n == 2:
                ax.legend(handles=handles, loc=loc)
            else:
                plt.gca().legend(handles=handles, loc=loc)

        # Adjust Axes
        seed_lims = [np.array(s.geometry.limits).flatten() for s in self]
        mins = np.array(seed_lims)[:, 0::2].min(axis=0)
        maxs = np.array(seed_lims)[:, 1::2].max(axis=0)
        xlim = (min(xlim[0], mins[0]), max(xlim[1], maxs[0]))
        ylim = (min(ylim[0], mins[1]), max(ylim[1], maxs[1]))
        if n == 2:
            plt.axis('square')
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
        if n == 3:
            zlim = (min(zlim[0], mins[2]), max(zlim[1], maxs[2]))
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_zlim(zlim)
            _misc.axisEqual3D(ax)

    # ----------------------------------------------------------------------- #
    # Position Function                                                       #
    # ----------------------------------------------------------------------- #
    def position(self, domain, pos_dists={}, rng_seed=0, hold=[],
                 max_attempts=10000, rtol='fit', verbose=False):
        """Position seeds in a domain

        This method positions the seeds within a domain. The "domain" should be
        a geometry instance from the :mod:`microstructpy.geometry` module.

        The "pos_dist" input is for phases with custom position distributions,
        the default being a uniform random distribution.
        For example:

        .. code-block:: python

            import scipy.stats
            mu = [0.5, -0.2]
            sigma = [[2.0, 0.3], [0.3, 0.5]]
            pos_dists = {2: scipy.stats.multivariate_normal(mu, sigma),
                         3: ['random',
                             scipy.stats.norm(0, 1)]
                         }

        Here, phases 0 and 1 have the default distribution, phase 2 has a
        bivariate normal position distribution, and phase 3 is uniform in the
        x and normally distributed in the y. Multivariate distributions are
        described in the multivariate section of the :mod:`scipy.stats`
        documentation.

        The position of certain seeds can be held fixed during the positioning
        process using the "hold" input. This should be a list of booleans,
        where False indicates a seed should not be held fixed and True
        indicates that it should be held fixed. The default behavior is to not
        hold any seeds fixed.

        The "rtol" parameter governs the relative overlap tolerable between
        seeds. Setting rtol to 0 means that there is no overlap, while a value
        of 1 means that one seed's center is on the edge of another seed.
        The default value is 'fit', which determines a tolerance between 0 and
        1 based on the ratio of standard deviation to mean in grain volumes.

        Args:
            domain (from :mod:`microstructpy.geometry`): The domain of the
                microstructure.
            pos_dists (dict): *(optional)* Position distributions for each
                phase, formatted like the example above.
                Defaults to uniform random throughout the domain.
            rng_seed (int): *(optional)* Random number generator (RNG) seed
                for positioning the seeds. Should be a non-negative integer.
            hold (list or numpy.ndarray): *(optional)* List of booleans for
                holding the positions of seeds.
                Defaults to False for all seeds.
            max_attempts (int): *(optional)* Number of random trials before
                removing a seed from the list.
                Defaults to 10,000.
            rtol (str or float): *(optional)* The relative overlap tolerance
                between seeds. This parameter should be between 0 and 1.
                Using the 'fit' option, a function will determine the value
                for rtol based on the mean and standard deviation in seed
                volumes.
            verbose (bool): *(optional)* This option will print a running
                counter of how many seeds have been positioned.
                Defaults to False.

        """  # NOQA: E501
        if len(hold) == 0:
            hold = [False for seed in self]

        # set the spatial distributions
        u_dist = [scipy.stats.uniform(lb, ub - lb) for lb, ub in
                  domain.sample_limits]

        distribs = []
        n_phases = max([s.phase for s in self]) + 1
        for i in range(n_phases):
            distribs.append(pos_dists.get(i, u_dist))

        # Add hold seeds
        n_seeds = len(self)
        tree = aabbtree.AABBTree()
        for i in range(n_seeds):
            if hold[i]:
                # add to tree
                aabb = aabbtree.AABB(self[i].geometry.limits)
                tree.add(aabb, i)

        positioned = np.array(hold)
        vols = np.array([s.volume for s in self])
        i_sort = np.flip(np.argsort(vols))
        posd_sort = positioned[i_sort]
        i_position = i_sort[~posd_sort]

        # allowable overlap, relative to radius
        cv = scipy.stats.variation(vols)
        if domain.n_dim == 2 and rtol == 'fit':
            numer = 0.362954 * cv * cv - 0.419069 * cv + .184959
            denom = cv * cv - 1.05989 * cv + 0.365096
            rtol = numer / denom
        elif rtol == 'fit':
            numer = 0.471115 * cv * cv - 0.602324 * cv + 0.297562
            denom = cv * cv - 1.08469 * cv + 0.428216
            rtol = numer / denom

        # position the remaining seeds
        i_reject = []
        np.random.seed(rng_seed)
        n_samples = 100

        for k, i in enumerate(i_position):
            if verbose:
                print(k + 1, 'of', len(i_position))

            seed = self[i]
            pos_dist = distribs[seed.phase]

            searching = True
            n_attempts = 0
            i_sample = 0
            pts = sample_pos_within(pos_dist, n_samples, domain)
            while searching and n_attempts < max_attempts:
                pt = pts[i_sample]

                seed.position = pt
                n_attempts += 1
                i_sample += 1

                if i_sample == n_samples:
                    pts = sample_pos_within(pos_dist, n_samples, domain)
                    i_sample = 0

                bkdwn = np.array(seed.breakdown)
                cens = bkdwn[:, :-1]
                rads = bkdwn[:, -1].reshape(-1, 1)

                aabb = aabbtree.AABB(seed.geometry.limits)
                olap_inds = tree.overlap_values(aabb, method='BFS')
                olap_seeds = self[olap_inds]
                clears = True
                for olap_seed in olap_seeds:
                    o_bkdwn = np.array(olap_seed.breakdown)
                    o_cens = o_bkdwn[:, :-1]
                    o_rads = o_bkdwn[:, -1].reshape(1, -1)

                    if len(rads) > 1:
                        dists = distance.cdist(cens, o_cens)
                    else:
                        rel_pos = o_cens - cens
                        rp2 = rel_pos * rel_pos
                        dists = np.sqrt(np.sum(rp2, axis=1))
                    tol = rtol * np.minimum(rads, o_rads)
                    total_dists = dists + tol - rads - o_rads
                    if np.any(total_dists < 0):
                        clears = False
                        break

                searching = not clears

            if searching:
                i_reject.append(i)
            else:
                positioned[i] = True
                self[i] = seed

                # add to tree
                aabb = aabbtree.AABB(seed.geometry.limits)
                tree.add(aabb, i)

        keep_mask = np.array(n_seeds * [True])
        keep_mask[i_reject] = False

        if ~np.all(keep_mask):
            reject_seeds = self[~keep_mask]
            f = 'seed_position_reject.log'
            reject_seeds.write(f)

            w_str = 'Seeds were removed from the seed list during positioning.'
            w_str += ' Their data has beeen written to ' + f + ' and their'
            w_str += ' indices were ' + str(i_reject) + '.'
            warnings.warn(w_str, RuntimeWarning)

        self.seeds = self[keep_mask].seeds


def _get_n_dim(phases):
    n_dim = None
    for phase in phases:
        if 'shape' in phase:
            n_dim = geometry.factory(phase['shape']).n_dim
    if n_dim is None:
        e_str = 'Number of dimensions could not be determined from phase '
        e_str += 'shapes. Consider setting the shape of a phase, or'
        e_str += ' specifying the number of dimensions.'
        raise ValueError(e_str)
    return n_dim


def _set_sample_rng_seeds(phases, rng_seeds, maxint):
    rng_keys = list({k for p in phases for k in p} - set(_misc.gen_kws))
    rng_keys.extend(['fraction', 'phase'])

    n_keys = len(rng_keys)
    int_step = maxint / n_keys
    sample_seeds = {}
    for i, k in enumerate(rng_keys):
        rng_seed = int(rng_seeds.get(k, 0) + i * int_step)
        sample_seeds[k] = rng_seed % maxint
    return sample_seeds


def _calc_pop_fracs(n_dim, phases, sample_rng_seeds, max_int):
    # compute volume of each phase
    vol_rng = sample_rng_seeds['fraction']
    n_phases = len(phases)
    rel_vols = np.ones(n_phases)
    for i, phase in enumerate(phases):
        vol = phase.get('fraction', 1)
        try:
            v_sample = -1
            while v_sample < 0:
                v_sample = vol.rvs(random_state=vol_rng)
                vol_rng = (vol_rng + 1) % max_int
            rel_vols[i] = v_sample
        except AttributeError:
            rel_vols[i] = vol
    vol_fracs = rel_vols / sum(rel_vols)

    # Compute the average grain volume of each phase
    if n_dim == 2:
        shape0 = geometry.factory(phases[0]['shape'])
        avg_vols = [geometry.factory(p['shape']).area_expectation(**p)
                    for p in phases]
    else:
        avg_vols = [geometry.factory(p['shape']).volume_expectation(**p)
                    for p in phases]
    weights = vol_fracs / np.array(avg_vols)
    pop_fracs = weights / sum(weights)
    return pop_fracs


def _sample_phase_args(phase, sample_rng_seeds, n_dim, maxint):
    seed_kwargs = {}
    for kw in set(phase) - set(_misc.gen_kws):
        rng_seed = sample_rng_seeds[kw]

        # Sample, with special cases for orientation
        if kw not in _misc.ori_kws:
            try:
                val = phase[kw].rvs(random_state=rng_seed)
            except AttributeError:
                val = phase[kw]
            seed_kwargs[kw] = val
        elif (phase[kw] == 'random') and (n_dim == 2):
            np.random.seed(rng_seed)
            ang_dist = scipy.stats.uniform(loc=0, scale=360)
            seed_kwargs['angle_deg'] = ang_dist.rvs(random_state=rng_seed)
        elif phase[kw] == 'random':
            quat_dist = scipy.stats.norm()
            elems = quat_dist.rvs(4, random_state=rng_seed)
            mag = np.linalg.norm(elems)
            elems /= mag
            val = Quaternion(elems).rotation_matrix
            seed_kwargs[kw] = val
        elif kw in ['rot_seq', 'rot_seq_deg', 'rot_seq_rad']:
            seq = []
            val = phase[kw]
            if not isinstance(val, list):
                val = [val]
            for rot_i, rotation in enumerate(val):
                rot_dict = {str(kw): rotation[kw] for kw in rotation}
                ax = rot_dict.get('axis', 'x')
                ang_dist = rot_dict.get('angle', 0)
                rot_rng = (rng_seed + rot_i) % maxint
                try:
                    ang = ang_dist.rvs(random_state=rot_rng)
                except AttributeError:
                    ang = ang_dist
                seq.append((ax, ang))
            seed_kwargs[kw] = seq
        else:
            try:
                val = phase[kw].rvs(random_state=rng_seed)
            except AttributeError:
                val = phase[kw]
            seed_kwargs[kw] = val

        # Update the RNG seed
        sample_rng_seeds[kw] = (rng_seed + 1) % maxint
    return seed_kwargs


def _plt_args(seeds, index_by, kwargs):
    seed_args = [{} for seed in seeds]
    for seed_num, seed in enumerate(seeds):
        phase_num = seed.phase
        for key, val in kwargs.items():
            if type(val) in (list, np.array):
                if index_by == 'seed' and len(val) > seed_num:
                    seed_args[seed_num][key] = val[seed_num]
                elif index_by == 'material' and len(val) > phase_num:
                    seed_args[seed_num][key] = val[phase_num]
            else:
                seed_args[seed_num][key] = val
    return seed_args


def _get_axes(n):
    if n == 2:
        ax = plt.gca()
    else:
        ax = plt.gcf().gca(projection=Axes3D.name)
    n_obj = _misc.ax_objects(ax)
    if n_obj > 0:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
    else:
        xlim = [float('inf'), -float('inf')]
        ylim = [float('inf'), -float('inf')]

    lims = [xlim, ylim]

    if n == 3:
        if n_obj > 0:
            zlim = ax.get_zlim()
        else:
            zlim = [float('inf'), -float('inf')]
        lims.append(zlim)
    return ax, lims


def _plot_2d(ax, seeds, seed_args):
    ellipse_data = {'w': [], 'h': [], 'a': [], 'xy': []}
    ec_kwargs = {}

    rect_data = []
    rect_kwargs = {}
    for seed, args in zip(seeds, seed_args):
        geom_name = type(seed.geometry).__name__.lower().strip()
        if geom_name == 'ellipse':
            ellipse_data['w'].append(2 * seed.geometry.a)
            ellipse_data['h'].append(2 * seed.geometry.b)
            ellipse_data['a'].append(seed.geometry.angle_deg)
            ellipse_data['xy'].append(np.array(seed.position))

            for key, val in args.items():
                val_list = ec_kwargs.get(key, []) + [val]
                ec_kwargs[key] = val_list

        elif geom_name == 'circle':
            ellipse_data['w'].append(seed.geometry.diameter)
            ellipse_data['h'].append(seed.geometry.diameter)
            ellipse_data['a'].append(0)
            ellipse_data['xy'].append(np.array(seed.position))

            for key, val in args.items():
                val_list = ec_kwargs.get(key, []) + [val]
                ec_kwargs[key] = val_list

        elif geom_name in ['rectangle', 'square']:
            w, h = seed.geometry.side_lengths
            corner = seed.geometry.corner
            t = seed.geometry.angle_deg
            rect_inputs = {'width': w, 'height': h, 'angle': t, 'xy': corner}
            rect_data.append(rect_inputs)

            for key, val in args.items():
                val_list = rect_kwargs.get(key, []) + [val]
                rect_kwargs[key] = val_list

        elif geom_name == 'nonetype':
            pass

        else:
            e_str = 'Cannot plot groups of ' + geom_name + ' yet.'
            raise NotImplementedError(e_str)

    # abbreviate kwargs if all the same
    for key, val in ec_kwargs.items():
        v1 = val[0]
        same = True
        for v in val:
            same &= v == v1
        if same:
            ec_kwargs[key] = v1

    # Plot Circles and Ellipses
    w = np.array(ellipse_data['w'])
    h = np.array(ellipse_data['h'])
    a = np.array(ellipse_data['a'])
    xy = np.array(ellipse_data['xy'])
    ec = collections.EllipseCollection(w, h, a, units='x', offsets=xy,
                                       transOffset=ax.transData, **ec_kwargs)
    ax.add_collection(ec)

    # Plot Rectangles
    rects = [Rectangle(**rect_inputs) for rect_inputs in rect_data]
    rc = collections.PatchCollection(rects, False, **rect_kwargs)
    ax.add_collection(rc)

    ax.autoscale_view()


def _plot_2d_breakdowns(ax, seeds, seed_args):
    breakdowns = np.zeros((0, 3))
    ec_kwargs = {}
    for seed, args in zip(seeds, seed_args):
        breakdowns = np.concatenate((breakdowns, seed.breakdown))
        n_c = len(seed.breakdown)
        for key, val in args.items():
            val_list = ec_kwargs.get(key, [])
            val_list.extend(n_c * [val])
            ec_kwargs[key] = val_list
    d = 2 * breakdowns[:, -1]
    xy = breakdowns[:, :-1]
    a = np.full(len(breakdowns), 0)

    # abbreviate kwargs if all the same
    for key, val in ec_kwargs.items():
        v1 = val[0]
        same = True
        for v in val:
            same &= v == v1
        if same:
            ec_kwargs[key] = v1

    ec = collections.EllipseCollection(d, d, a, units='x', offsets=xy,
                                       transOffset=ax.transData, **ec_kwargs)
    ax.add_collection(ec)
    ax.autoscale_view()


def _add_legend(ax, material, seeds, seed_args, kwargs, index_by, loc):
    if material:
        p_kwargs = [{'label': m} for m in material]
        if index_by == 'seed':
            for seed_kwargs, seed in zip(seed_args, seeds):
                p_kwargs[seed.phase].update(seed_kwargs)
        else:
            for key, val in kwargs.items():
                if type(val) in (list, np.array):
                    for i, elem in enumerate(val):
                        p_kwargs[i][key] = elem
                else:
                    for p_kwargs_i in p_kwargs:
                        p_kwargs_i[key] = val

        # Replace plural keywords
        for p_kw in p_kwargs:
            for kw in _misc.mpl_plural_kwargs:
                if kw in p_kw:
                    p_kw[kw[:-1]] = p_kw[kw]
                    del p_kw[kw]
        ax.legend(handles=[patches.Patch(**p_kw) for p_kw in p_kwargs],
                  loc=loc)


def calc_rtol(seeds):
    """Calculate relative overlap tolerance."""
    cv = scipy.stats.variation([s.volume for s in seeds])
    n_dim = seeds[0].geometry.n_dim
    if n_dim == 2:
        numer = 0.362954 * cv * cv - 0.419069 * cv + .184959
        denom = cv * cv - 1.05989 * cv + 0.365096
        rtol = numer / denom
    elif n_dim == 3:
        numer = 0.471115 * cv * cv - 0.602324 * cv + 0.297562
        denom = cv * cv - 1.08469 * cv + 0.428216
        rtol = numer / denom
    else:
        raise ValueError('Cannot calculate rtol for {}-D.'.format(n_dim))
    return rtol


def sample_pos(distribution, n=1):
    """ Sample position distribution

    This function returns a sample of the postion distribution.
    This distribution can be either a list of independent distributions
    for each axis, or a single multi-variate distribution. A list of
    multi-variate distributions is given on the `SciPy stats website`_.

    Two examples of position distributions are given below.

    .. code-block:: python

        # three independent distributions
        distribution = [scipy.stats.uniform(-1, 2),
                        scipy.stats.norm(0, 1),
                        scipy.stats.binom(5, 0.4)]

        # one multi-variate distribution
        mu = [2, -3 , 5]
        sigma = [[1, 3, 0], [3, 1, 2], [0, 2, 2]]
        distribution = scipy.stats.multivariate_normal(mu, sigma)

    Args:
        distribution (list or scipy.stats distribution): The position
            distribution.

        n (int): *(optional)* Number of samples. Defaults to 1.

    Returns:
        list: A sample of the distribution.
    """  # NOQA : E501
    try:
        pos = distribution.rvs(n)
    except AttributeError:
        pos = np.full((n, len(distribution)), 0, dtype='float')
        for j, coord_dist in enumerate(distribution):
            try:
                pos[:, j] = coord_dist.rvs(n)
            except AttributeError:
                pos[:, j] = coord_dist

    if n == 1:
        return pos[0]
    else:
        return pos


def sample_pos_within(distribution, n, domain):
    pos = []
    while len(pos) < n:
        samples = sample_pos(distribution, n)
        mask = domain.within(samples)
        pos.extend(samples[mask])
    if n == 1:
        return pos
    return np.array(pos[:n])



import microstructpy as msp

c1 = msp.geometry.n_sphere.NSphere(r=1, center=(0, 4, -1))
c2 = msp.geometry.n_sphere.NSphere(r=2, center=(3, 1))
c3 = msp.geometry.n_sphere.NSphere(d=4, center=(3, 1))


def test_eq():
    assert c1 == c1
    assert not c2 == c1

    assert c2 == c3
